# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from typing import List
from typing import Literal
from typing import Union

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.backend import standardize_schema
from synalinks.src.backend.common.json_schema_utils import concatenate_schema
from synalinks.src.backend.common.json_schema_utils import contains_schema
from synalinks.src.backend.common.json_schema_utils import factorize_schema
from synalinks.src.backend.common.json_schema_utils import in_mask_schema
from synalinks.src.backend.common.json_schema_utils import is_schema_equal
from synalinks.src.backend.common.json_schema_utils import out_mask_schema


class JsonSchemaConcatenateTest(testing.TestCase):
    def test_concatenate_identical_schemas(self):
        class Input(DataModel):
            foo: str

        class Result(DataModel):
            foo: str
            foo_1: str

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = concatenate_schema(schema, schema)
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_concatenate_schemas_with_different_properties(self):
        class Input1(DataModel):
            foo: str

        class Input2(DataModel):
            bar: str

        class Result(DataModel):
            foo: str
            bar: str

        schema1 = standardize_schema(Input1.get_schema())
        schema2 = standardize_schema(Input2.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = concatenate_schema(schema1, schema2)
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_concatenate_similar_entities(self):
        class City(DataModel):
            label: Literal["City"]
            name: str

        class Cities(DataModel):
            entities: List[City]

        class Result(DataModel):
            entities: List[City]
            entities_1: List[City]

        schema1 = standardize_schema(Cities.get_schema())
        schema2 = standardize_schema(Cities.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = concatenate_schema(schema1, schema2)
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_concatenate_different_entities(self):
        class City(DataModel):
            label: Literal["City"]
            name: str

        class Event(DataModel):
            label: Literal["Event"]
            name: str

        class Cities(DataModel):
            entities: List[City]

        class Events(DataModel):
            entities: List[Event]

        class Result(DataModel):
            entities: List[City]
            entities_1: List[Event]

        schema1 = standardize_schema(Cities.get_schema())
        schema2 = standardize_schema(Events.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = concatenate_schema(schema1, schema2)
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_concatenate_schema_multiple_times(self):
        class Input(DataModel):
            foo: str

        class Result(DataModel):
            foo: str
            foo_1: str
            foo_2: str

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = concatenate_schema(schema, schema)
        result_schema = concatenate_schema(result_schema, schema)
        self.assertTrue(is_schema_equal(result_schema, expected_schema))


class JsonSchemaFactorizeTest(testing.TestCase):
    def test_factorize_schema_with_identical_properties(self):
        class Input(DataModel):
            foo: str
            foo_1: str

        class Result(DataModel):
            foos: List[str]

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = factorize_schema(schema)
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_factorize_schema_with_multiple_identical_properties(self):
        class Input(DataModel):
            foo: str
            foo_1: str
            foo_2: str

        class Result(DataModel):
            foos: List[str]

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = factorize_schema(schema)
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_factorize_schema_with_different_properties(self):
        class Input(DataModel):
            foo: str
            foo_1: str
            foo_2: str

        class Result(DataModel):
            foos: List[str]

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = factorize_schema(schema)
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_factorize_schema_with_mixed_properties(self):
        class Input(DataModel):
            foo: str
            foo_1: str
            bar: str

        class Result(DataModel):
            foos: List[str]
            bar: str

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = factorize_schema(schema)
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_factorize_schema_with_existing_array_property(self):
        class Input(DataModel):
            foos: List[str]
            foo: str

        class Result(DataModel):
            foos: List[str]

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = factorize_schema(schema)
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_factorize_schema_with_existing_array_property_and_additional_properties(
        self,
    ):
        class Input(DataModel):
            foos: List[str]
            foo: str
            foo_1: str

        class Result(DataModel):
            foos: List[str]

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = factorize_schema(schema)
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_factorize_schema_with_multiple_groups_of_properties(self):
        class Input(DataModel):
            foo: str
            foo_1: str
            bar: str
            bar_1: str

        class Result(DataModel):
            foos: List[str]
            bars: List[str]

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = factorize_schema(schema)
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_factorize_similar_entities(self):
        class City(DataModel):
            label: Literal["City"]
            name: str

        class Input(DataModel):
            entities: List[City]
            entities_1: List[City]

        class Result(DataModel):
            entities: List[City]

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = factorize_schema(schema)
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_factorize_different_entities(self):
        class City(DataModel):
            label: Literal["City"]
            name: str

        class Event(DataModel):
            label: Literal["Event"]
            name: str

        class Input(DataModel):
            entities: List[City]
            entities_1: List[Event]

        class Result(DataModel):
            entities: List[Union[City, Event]]

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = factorize_schema(schema)
        self.assertTrue(is_schema_equal(result_schema, expected_schema))


class JsonSchemaOutMaskTest(testing.TestCase):
    def test_mask_basic(self):
        class Input(DataModel):
            foo: str
            bar: str

        class Result(DataModel):
            bar: str

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = out_mask_schema(schema, mask=["foo"])
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_mask_multiple_fields_with_same_base_name(self):
        class Input(DataModel):
            foo: str
            foo_1: str
            bar: str
            bar_1: str

        class Result(DataModel):
            bar: str
            bar_1: str

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = out_mask_schema(schema, mask=["foo"])
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_mask_nested(self):
        class BarObject(DataModel):
            foo: str
            bar: str

        class Input(DataModel):
            foo: str
            foo_1: str
            bar: BarObject

        schema = standardize_schema(Input.get_schema())

        class BarObject(DataModel):
            bar: str

        class Result(DataModel):
            bar: BarObject

        expected_schema = standardize_schema(Result.get_schema())
        result_schema = out_mask_schema(schema, mask=["foo"])
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_mask_deeply_nested(self):
        class BooObject(DataModel):
            foo: str
            boo: str

        class BarObject(DataModel):
            boo: BooObject

        class Input(DataModel):
            foo: str
            bar: BarObject

        schema = standardize_schema(Input.get_schema())

        class BooObject(DataModel):
            boo: str

        class BarObject(DataModel):
            boo: BooObject

        class Result(DataModel):
            bar: BarObject

        expected_schema = standardize_schema(Result.get_schema())
        result_schema = out_mask_schema(schema, mask=["foo"])
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_mask_array(self):
        class BarObject(DataModel):
            foo: str
            bar: str

        class Input(DataModel):
            bars: List[BarObject]

        schema = standardize_schema(Input.get_schema())

        class BarObject(DataModel):
            bar: str

        class Result(DataModel):
            bars: List[BarObject]

        expected_schema = standardize_schema(Result.get_schema())

        result_schema = out_mask_schema(schema, mask=["foo"])
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_mask_empty_schema(self):
        class Input(DataModel):
            pass

        class Result(DataModel):
            pass

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = out_mask_schema(schema, mask=["foo"])
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_empty_mask_list(self):
        class Input(DataModel):
            foo: str
            bar: str

        class Result(DataModel):
            foo: str
            bar: str

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = out_mask_schema(schema, mask=[])
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_mask_non_recursive(self):
        class BarObject(DataModel):
            foo: str
            bar: str

        class Input(DataModel):
            foo: str
            foo_1: str
            bar: BarObject

        class Result(DataModel):
            bar: BarObject

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = out_mask_schema(schema, mask=["foo"], recursive=False)
        self.assertTrue(is_schema_equal(result_schema, expected_schema))


class JsonSchemaInMaskTest(testing.TestCase):
    def test_mask_basic(self):
        class Input(DataModel):
            foo: str
            bar: str

        class Result(DataModel):
            foo: str

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = in_mask_schema(schema, mask=["foo"])
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_mask_multiple_fields_with_same_base_name(self):
        class Input(DataModel):
            foo: str
            foo_1: str
            bar: str
            bar_1: str

        class Result(DataModel):
            foo: str
            foo_1: str

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = in_mask_schema(schema, mask=["foo"])
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_mask_nested(self):
        class BarObject(DataModel):
            foo: str
            bar: str
            boo: str

        class Input(DataModel):
            foo: str
            foo_1: str
            bar: BarObject
            boo: str

        schema = standardize_schema(Input.get_schema())

        class BarObject(DataModel):
            foo: str
            bar: str

        class Result(DataModel):
            foo: str
            foo_1: str
            bar: BarObject

        expected_schema = standardize_schema(Result.get_schema())
        result_schema = in_mask_schema(schema, mask=["foo", "bar"])
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_mask_deeply_nested(self):
        class BooObject(DataModel):
            foo: str
            boo: str

        class BarObject(DataModel):
            boo: BooObject

        class Input(DataModel):
            foo: str
            bar: BarObject

        schema = standardize_schema(Input.get_schema())

        class BooObject(DataModel):
            foo: str

        class BarObject(DataModel):
            pass

        class Result(DataModel):
            foo: str
            bar: BarObject

        expected_schema = standardize_schema(Result.get_schema())
        result_schema = in_mask_schema(schema, mask=["foo", "bar"])
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_mask_array(self):
        class BarObject(DataModel):
            foo: str
            bar: str

        class Input(DataModel):
            bars: List[BarObject]

        schema = standardize_schema(Input.get_schema())

        class BarObject(DataModel):
            bar: str

        class Result(DataModel):
            bars: List[BarObject]

        expected_schema = standardize_schema(Result.get_schema())

        result_schema = in_mask_schema(schema, mask=["bar"])
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_mask_empty_schema(self):
        class Input(DataModel):
            pass

        class Result(DataModel):
            pass

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = in_mask_schema(schema, mask=["foo"])
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_mask_empty_mask_list(self):
        class Input(DataModel):
            foo: str
            bar: str

        class Result(DataModel):
            pass

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = in_mask_schema(schema, mask=[])
        self.assertTrue(is_schema_equal(result_schema, expected_schema))

    def test_mask_non_recursive(self):
        class BarObject(DataModel):
            foo: str
            bar: str
            boo: str

        class Input(DataModel):
            foo: str
            foo_1: str
            bar: BarObject
            boo: str

        class Result(DataModel):
            foo: str
            foo_1: str
            bar: BarObject

        schema = standardize_schema(Input.get_schema())
        expected_schema = standardize_schema(Result.get_schema())

        result_schema = in_mask_schema(schema, mask=["foo", "bar"], recursive=False)
        self.assertTrue(is_schema_equal(result_schema, expected_schema))


class JsonSchemaContainsTest(testing.TestCase):
    def test_contains_same_schema(self):
        class Input1(DataModel):
            foo: str
            bar: str

        class Input2(DataModel):
            foo: str
            bar: str

        schema1 = standardize_schema(Input1.get_schema())
        schema2 = standardize_schema(Input2.get_schema())

        self.assertTrue(contains_schema(schema1, schema2))

    def test_contains_subset_schema(self):
        class Input1(DataModel):
            foo: str
            bar: str

        class Input2(DataModel):
            foo: str

        schema1 = standardize_schema(Input1.get_schema())
        schema2 = standardize_schema(Input2.get_schema())

        self.assertTrue(contains_schema(schema1, schema2))

    def test_contains_different_schema(self):
        class Input1(DataModel):
            bar: str

        class Input2(DataModel):
            foo: str

        schema1 = standardize_schema(Input1.get_schema())
        schema2 = standardize_schema(Input2.get_schema())

        self.assertFalse(contains_schema(schema1, schema2))



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import json
from unittest.mock import patch

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.backend import Field
from synalinks.src.language_models import LanguageModel
from synalinks.src.modules.core.input_module import Input
from synalinks.src.modules.ttc.chain_of_thought import ChainOfThought
from synalinks.src.programs.program import Program


class ChainOfThoughtModuleTest(testing.TestCase):
    async def test_default_reasoning_effort_is_low(self):
        """Test that ChainOfThought defaults to 'low' reasoning effort."""

        class Answer(DataModel):
            answer: str = Field(description="The answer")

        language_model = LanguageModel(model="ollama/mistral")

        cot = ChainOfThought(
            data_model=Answer,
            language_model=language_model,
            name="test_cot",
        )

        self.assertEqual(cot.reasoning_effort, "low")

    async def test_explicit_reasoning_effort_none(self):
        """Test that reasoning_effort can be explicitly set to 'none'."""

        class Answer(DataModel):
            answer: str = Field(description="The answer")

        language_model = LanguageModel(model="ollama/mistral")

        cot = ChainOfThought(
            data_model=Answer,
            language_model=language_model,
            reasoning_effort="none",
            name="test_cot",
        )

        self.assertEqual(cot.reasoning_effort, "none")

    @patch("litellm.supports_reasoning")
    @patch("litellm.acompletion")
    async def test_thinking_generated_in_structured_output(
        self, mock_completion, mock_supports_reasoning
    ):
        """Test thinking field is generated by the model as part of structured output."""
        mock_supports_reasoning.return_value = False

        class Query(DataModel):
            query: str = Field(description="The user query")

        class Answer(DataModel):
            answer: str = Field(description="The correct answer")

        language_model = LanguageModel(model="openai/gpt-4o")

        x0 = Input(data_model=Query)
        x1 = await ChainOfThought(
            data_model=Answer,
            language_model=language_model,
            reasoning_effort="low",
        )(x0)

        program = Program(
            inputs=x0,
            outputs=x1,
            name="test_thinking_in_output",
            description="Test thinking generated in structured output",
        )

        # Thinking is always part of the schema and generated by the model
        expected_response = {
            "thinking": "The capital of France is a well-known fact.",
            "answer": "Paris",
        }

        mock_completion.return_value = {
            "choices": [{"message": {"content": json.dumps(expected_response)}}]
        }

        result = await program(Query(query="What is the capital of France?"))

        result_json = result.get_json()
        self.assertEqual(
            result_json["thinking"], "The capital of France is a well-known fact."
        )
        self.assertEqual(result_json["answer"], "Paris")

    @patch("litellm.acompletion")
    async def test_chain_of_thought_with_reasoning_effort_none(self, mock_completion):
        """Test ChainOfThought with reasoning_effort='none' (thinking in schema)."""

        class Query(DataModel):
            query: str = Field(
                description="The user query",
            )

        class Answer(DataModel):
            answer: str = Field(
                description="The correct answer",
            )

        language_model = LanguageModel(
            model="ollama/mistral",
        )

        x0 = Input(data_model=Query)
        x1 = await ChainOfThought(
            data_model=Answer,
            language_model=language_model,
            reasoning_effort="none",
        )(x0)

        program = Program(
            inputs=x0,
            outputs=x1,
            name="chain_of_tought_qa",
            description="Answer the user query step by step",
        )

        expected_string = (
            """{"thinking": "Toulouse hosts numerous research institutions """
            """and universities that specialize in aerospace engineering and """
            """robotics, such as the Institut Supérieur de l'Aéronautique et """
            """de l'Espace (ISAE-SUPAERO) and the French National Centre for """
            """Scientific Research (CNRS)","""
            """ "answer": "Toulouse"}"""
        )

        mock_completion.return_value = {
            "choices": [{"message": {"content": expected_string}}]
        }

        result = await program(
            Query(
                query="What is the French city of aerospace and robotics?",
            )
        )

        self.assertEqual(result.get_json(), json.loads(expected_string))



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from enum import Enum
from typing import List
from typing import Literal
from typing import Union

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.backend import Field
from synalinks.src.backend import is_schema_equal
from synalinks.src.backend.common.dynamic_json_schema_utils import dynamic_enum
from synalinks.src.backend.common.dynamic_json_schema_utils import dynamic_tool_calls
from synalinks.src.backend.common.dynamic_json_schema_utils import dynamic_tool_choice
from synalinks.src.utils.tool_utils import Tool


async def calculate(expression: str):
    """Calculate the result of a mathematical expression.

    Args:
        expression (str): The mathematical expression to calculate, such as
            '2 + 2'. The expression can contain numbers, operators (+, -, *, /),
            parentheses, and spaces.
    """
    if not all(char in "0123456789+-*/(). " for char in expression):
        return {
            "result": None,
            "log": "Error: invalid characters in expression",
        }
    try:
        # Evaluate the mathematical expression safely
        result = round(float(eval(expression, {"__builtins__": None}, {})), 2)
        return {
            "result": result,
            "log": "Successfully executed",
        }
    except Exception as e:
        return {
            "result": None,
            "log": f"Error: {e}",
        }


async def thinking(thinking: str):
    """Think about something.

    Args:
        thinking (str): Your step by step thinking.
    """
    return {
        "thinking": thinking,
    }


class DynamicEnumTest(testing.TestCase):
    def test_basic_dynamic_enum(self):
        class DecisionAnswer(DataModel):
            thinking: str
            choice: str

        class Choice(str, Enum):
            easy = "easy"
            difficult = "difficult"
            unknown = "unknown"

        class Decision(DataModel):
            thinking: str
            choice: Choice

        labels = ["easy", "difficult", "unkown"]

        schema = dynamic_enum(DecisionAnswer.get_schema(), "choice", labels)

        self.assertTrue(is_schema_equal(Decision.get_schema(), schema))


class DynamicToolCallsSchemaTest(testing.TestCase):
    def test_dynamic_tool_call_schema(self):
        class Calculate(DataModel):
            """Calculate the result of a mathematical expression."""

            tool_name: Literal["calculate"]
            expression: str = Field(
                description=(
                    "The mathematical expression to calculate, such as "
                    "'2 + 2'. The expression can contain numbers, operators (+, -, *, /),"
                    " parentheses, and spaces."
                )
            )

        class Thinking(DataModel):
            """Think about something."""

            tool_name: Literal["thinking"]
            thinking: str = Field(description="Your step by step thinking.")

        class ToolCalls(DataModel):
            tool_calls: List[Union[Calculate, Thinking]]

        expected_schema = ToolCalls.get_schema()

        tools = [
            Tool(calculate),
            Tool(thinking),
        ]

        dynamic_schema = dynamic_tool_calls(tools=tools)

        print("Expected:")
        print(ToolCalls.prettify_schema())
        print("Generated:")
        import json

        print(json.dumps(dynamic_schema, indent=2))

        self.assertEqual(expected_schema, dynamic_schema)


class DynamicToolChoiceSchemaTest(testing.TestCase):
    def test_dynamic_tool_call_schema(self):
        class Calculate(DataModel):
            """Calculate the result of a mathematical expression."""

            tool_name: Literal["calculate"]
            expression: str = Field(
                description=(
                    "The mathematical expression to calculate, such as "
                    "'2 + 2'. The expression can contain numbers, operators (+, -, *, /),"
                    " parentheses, and spaces."
                )
            )

        class Thinking(DataModel):
            """Think about something."""

            tool_name: Literal["thinking"]
            thinking: str = Field(description="Your step by step thinking.")

        class ToolChoice(DataModel):
            tool_choice: Union[Calculate, Thinking]

        expected_schema = ToolChoice.get_schema()

        tools = [
            Tool(calculate),
            Tool(thinking),
        ]

        dynamic_schema = dynamic_tool_choice(tools=tools)
        print("Expected:")
        print(ToolChoice.prettify_schema())
        print("Generated:")
        import json

        print(json.dumps(dynamic_schema, indent=2))

        self.assertEqual(expected_schema, dynamic_schema)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import copy
import inspect

from synalinks.src.api_export import synalinks_export
from synalinks.src.backend.common.json_schema_utils import standardize_schema
from synalinks.src.backend.common.symbolic_data_model import SymbolicDataModel
from synalinks.src.utils.async_utils import run_maybe_nested
from synalinks.src.utils.naming import auto_name


@synalinks_export("synalinks.JsonDataModel")
class JsonDataModel:
    """A backend-independent dynamic data model.

    This structure is the one flowing in the pipelines as
    the backend data models are only used for the variable/data model declaration.

    Args:
        schema (dict): The JSON object's schema. If not provided,
            uses the data model to infer it.
        json (dict): The JSON object's json. If not provided,
            uses the data model to infer it.
        data_model (DataModel | JsonDataModel): The data model to use to
            infer the schema and json.
        name (str): Optional. The name of the data model, automatically
            inferred if not provided.

    Examples:

    **Creating a `JsonDataModel` with a DataModel's schema and json:**

    ```python
    class Query(synalinks.DataModel):
        query: str = synalinks.Field(
            description="The user query",
        )

    json = {"query": "What is the capital of France?"}

    data_model = JsonDataModel(
        schema=Query.get_schema(),
        json=json,
    )
    ```

    **Creating a `JsonDataModel` with a data_model:**

    ```python
    class Query(synalinks.DataModel):
        query: str = synalinks.Field(
            description="The user query",
        )

    query_instance = Query(
        query="What is the capital of France?"
    )
    data_model = JsonDataModel(
        data_model=query_instance,
    )
    ```

    **Creating a `JsonDataModel` with `to_json_data_model()`:**

    ```python
    class Query(synalinks.DataModel):
        query: str = synalinks.Field(
            description="The user query",
        )

    data_model = Query(
        query="What is the capital of France?",
    ).to_json_data_model()
    ```
    """

    def __init__(
        self,
        schema=None,
        json=None,
        data_model=None,
        name=None,
    ):
        name = name or auto_name(self.__class__.__name__)
        self.name = name
        self._schema = None
        self._json = None

        if not data_model and not schema and not json:
            raise ValueError("Initializing without arguments is not permited.")
        if not schema and not data_model:
            raise ValueError(
                "You should specify at least one argument between "
                "`data_model` or `schema`."
            )
        if not schema and not json and not data_model:
            raise ValueError(
                "You should specify at least one argument between `data_model` or `json`."
            )
        if data_model:
            if not schema:
                schema = data_model.get_schema()
            if not json:
                if inspect.isclass(data_model):
                    raise ValueError(
                        "Couldn't get the JSON data from the `data_model` argument, "
                        "the `data_model` needs to be instanciated. "
                        f"Received data_model={data_model}."
                    )
                json = data_model.get_json()

        self._schema = standardize_schema(schema)
        self._json = json

    def to_symbolic_data_model(self):
        """Converts the JsonDataModel to a SymbolicDataModel.

        Returns:
            (SymbolicDataModel): The symbolic data model.
        """
        return SymbolicDataModel(schema=self._schema)

    def get_json(self):
        """Gets the current json of the JSON object.

        Returns:
            (dict): The current json of the JSON object.
        """
        return self._json

    def get_schema(self):
        """Gets the schema of the JSON object.

        Returns:
            (dict): The JSON schema.
        """
        return self._schema

    def prettify_schema(self):
        """Get a pretty version of the JSON schema for display.

        Returns:
            (dict): The indented JSON schema.
        """
        import orjson

        return orjson.dumps(self._schema, option=orjson.OPT_INDENT_2).decode()

    def prettify_json(self):
        """Get a pretty version of the JSON object for display.

        Returns:
            (str): The indented JSON object.
        """
        import orjson

        return orjson.dumps(self._json, option=orjson.OPT_INDENT_2).decode()

    def __add__(self, other):
        """Concatenates this data model with another.

        Args:
            other (JsonDataModel | DataModel):
                The other data model to concatenate with.

        Returns:
            (JsonDataModel): The concatenated data model.
        """
        from synalinks.src import ops

        return run_maybe_nested(
            ops.Concat().call(self, other),
        )

    def __radd__(self, other):
        """Concatenates another data model with this one.

        Args:
            other (JsonDataModel | DataModel):
                The other data model to concatenate with.

        Returns:
            (JsonDataModel): The concatenated data model.
        """
        from synalinks.src import ops

        return run_maybe_nested(
            ops.Concat().call(other, self),
        )

    def __and__(self, other):
        """Perform a `logical_and` with another data model.

        If one of them is None, output None. If both are provided,
        then concatenates this data model with the other.

        Args:
            other (JsonDataModel | DataModel): The other data model to concatenate with.

        Returns:
            (JsonDataModel | None): The concatenated data model or None
                based on the `logical_and` table.
        """
        from synalinks.src import ops

        return run_maybe_nested(
            ops.And().call(self, other),
        )

    def __rand__(self, other):
        """Perform a `logical_and` (reverse) with another data model.

        If one of them is None, output None. If both are provided,
        then concatenates the other data model with this one.

        Args:
            other (JsonDataModel | DataModel): The other data model to concatenate with.

        Returns:
            (JsonDataModel | None): The concatenated data model or None
                based on the `logical_and` table.
        """
        from synalinks.src import ops

        return run_maybe_nested(
            ops.And().call(other, self),
        )

    def __or__(self, other):
        """Perform a `logical_or` with another data model

        If one of them is None, output the other one. If both are provided,
        then concatenates this data model with the other.

        Args:
            other (JsonDataModel | DataModel): The other data model to concatenate with.

        Returns:
            (JsonDataModel | None): The concatenation of data model if both are provided,
                or the non-None data model or None if none are provided.
                (See `logical_or` table).
        """
        from synalinks.src import ops

        return run_maybe_nested(
            ops.Or().call(self, other),
        )

    def __ror__(self, other):
        """Perform a `logical_or` (reverse) with another data model

        If one of them is None, output the other one. If both are provided,
        then concatenates the other data model with this one.

        Args:
            other (JsonDataModel | DataModel): The other data model to concatenate with.

        Returns:
            (JsonDataModel | None): The concatenation of data model if both are provided,
                or the non-None data model or None if none are provided.
                (See `logical_or` table).
        """
        from synalinks.src import ops

        return run_maybe_nested(
            ops.Or().call(other, self),
        )

    def __xor__(self, other):
        """Perform a `logical_xor` with another data model.

        If one of them is `None`, output the other one. If both are provided,
        then the output is `None`.

        Args:
            other (SymbolicDataModel): The other data model to concatenate with.

        Returns:
            (SymbolicDataModel | None): `None` if both are
                provided, or the non-None data model if one is provided
                or `None` if none are provided. (See `logical_xor` table).
        """
        from synalinks.src import ops

        return run_maybe_nested(
            ops.Xor().call(self, other),
        )

    def __rxor__(self, other):
        """Perform a `logical_xor` (reverse) with another data model.

        If one of them is None, output the other one. If both are provided,
        then concatenates the other data model with this one.

        Args:
            other (SymbolicDataModel | DataModel): The other data model to concatenate
                with.

        Returns:
            (SymbolicDataModel | None): `None` if both are
                provided, or the non-None data model if one is provided
                or `None` if none are provided. (See `logical_xor` table).
        """
        from synalinks.src import ops

        return run_maybe_nested(
            ops.Xor().call(other, self),
        )

    def __contains__(self, other):
        """Check if the schema of `other` is contained in this one,
        or if a string key exists.

        Args:
            other (SymbolicDataModel | DataModel | str): The other data model to compare
                with, or a string key to check for in the schema properties.

        Returns:
            (bool): True if all properties of `other` are present in this one,
                or if the string key exists in the schema properties.
        """
        if isinstance(other, str):
            schema = self.get_schema()
            return other in schema.get("properties", {})
        from synalinks.src.backend.common.json_schema_utils import contains_schema

        return contains_schema(self.get_schema(), other.get_schema())

    def factorize(self):
        """Factorizes the data model.

        Returns:
            (JsonDataModel): The factorized data model.
        """
        from synalinks.src import ops

        return run_maybe_nested(
            ops.Factorize().call(self),
        )

    def in_mask(self, mask=None, recursive=True):
        """Applies a mask to **keep only** specified keys of the data model.

        Args:
            mask (list): The mask to be applied.
            recursive (bool): Optional. Whether to apply the mask recursively.
                Defaults to True.

        Returns:
            (JsonDataModel): The data model with the mask applied.
        """
        from synalinks.src import ops

        return run_maybe_nested(
            ops.InMask(mask=mask, recursive=recursive).call(self),
        )

    def out_mask(self, mask=None, recursive=True):
        """Applies a mask to **remove** specified keys of the data model.

        Args:
            mask (list): The mask to be applied.
            recursive (bool): Optional. Whether to apply the mask recursively.
                Defaults to True.

        Returns:
            (JsonDataModel): The data model with the mask applied.
        """
        from synalinks.src import ops

        return run_maybe_nested(
            ops.OutMask(mask=mask, recursive=recursive).call(self),
        )

    def __invert__(self):
        """Perform an invertion/negation

        When an input is provided, invert it by outputing `None`

        Returns:
            (SymbolicDataModel | None): `None` if used with an instance/class,
                and a symbolic data model if used on a metaclass or symbolic model.
        """
        from synalinks.src import ops

        return run_maybe_nested(
            ops.Not().call(self),
        )

    def prefix(self, prefix=None):
        """Add a prefix to **all** the data model fields (non-recursive).

        Args:
            prefix (str): the prefix to add.

        Returns:
            (JsonDataModel): The data model with the prefix added.
        """
        from synalinks.src import ops

        return run_maybe_nested(
            ops.Prefix(prefix=prefix).call(self),
        )

    def suffix(self, suffix=None):
        """Add a suffix to **all** the data model fields (non-recursive).

        Args:
            suffix (str): the suffix to add.

        Returns:
            (JsonDataModel): The data model with the suffix added.
        """
        from synalinks.src import ops

        return run_maybe_nested(
            ops.Suffix(suffix=suffix).call(self),
        )

    def get(self, key, default=None):
        """Get wrapper to make it easier to access JSON fields.

        Args:
            key (str): The key to access.
            default (any): The default value if key not found.
        """
        return self._json.get(key, default)

    def __getitem__(self, key):
        """Get item wrapper to make it easier to access JSON fields.

        Args:
            key (str): The key to access.
        """
        return self._json[key]

    def keys(self):
        """Keys wrapper to make it easier to access JSON fields."""
        return self._json.keys()

    def values(self):
        """Values wrapper to make it easier to access JSON fields."""
        return self._json.values()

    def items(self):
        """Items wrapper to make it easier to access JSON fields."""
        return self._json.items()

    def update(self, kv_dict):
        """Update wrapper to make it easier to modify JSON fields.

        Args:
            kv_dict (dict): The key/json dict to update.
        """
        self._json.update(kv_dict)

    def clone(self, name=None):
        """Clone a data model and give it a different name."""

        clone = copy.deepcopy(self)
        if name:
            clone.name = name
        else:
            clone.name = auto_name("clone_" + self.name)
        return clone

    def __repr__(self):
        return f"<JsonDataModel schema={self._schema}, json={self._json}>"


@synalinks_export(
    [
        "synalinks.utils.is_json_data_model",
        "synalinks.backend.is_json_data_model",
    ]
)
def is_json_data_model(x):
    """Returns whether `x` is a backend-independent data model.

    Args:
        x (any): The object to check.

    Returns:
        (bool): True if `x` is a backend-independent data model, False otherwise.
    """
    return isinstance(x, JsonDataModel)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import random

import numpy

from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import out_mask_json
from synalinks.src.optimizers.optimizer import Optimizer
from synalinks.src.saving import serialization_lib


@synalinks_export("synalinks.optimizers.EvolutionaryOptimizer")
class EvolutionaryOptimizer(Optimizer):
    """Abstract base class for evolutionary LLM-based optimizers.

    Evolutionary optimizers use LLMs for mutation and crossover operations,
    with configurable selection strategies for choosing candidates to evolve.

    Subclasses must implement:
        - `build()`: Build mutation and crossover programs
        - `mutate_candidate()`: Apply mutation to generate a new candidate
        - `merge_candidate()`: Apply crossover to merge two candidates

    Args:
        language_model (LanguageModel): The language model to use for
            mutation and crossover.
        mutation_temperature (float): The temperature for the LM calls of
            the mutation programs.
        crossover_temperature (float): The temperature for the LM calls of
            the crossover programs.
        selection (str): The method to select the candidate to evolve at the
            beginning of a batch between ['random', 'best', 'softmax'].
            (Default to 'softmax').
        selection_temperature (float): The temperature for softmax selection.
            Used only when `selection='softmax'`. Lower values concentrate
            selection on high-reward candidates, higher values make selection
            more uniform (Default 0.3).
        merging_rate (float): Rate at which crossover vs mutation is selected.
            (Default to 0.02).
        population_size (int): The maximum number of best candidates to keep
            during the optimization process.
        name (str): Optional name for the optimizer instance.
        description (str): Optional description of the optimizer instance.
    """

    def __init__(
        self,
        language_model=None,
        mutation_temperature=0.3,
        crossover_temperature=0.3,
        selection="softmax",
        selection_temperature=0.3,
        merging_rate=0.02,
        population_size=10,
        name=None,
        description=None,
        **kwargs,
    ):
        super().__init__(
            population_size=population_size,
            name=name,
            description=description,
        )
        self.merging_rate = merging_rate
        self.language_model = language_model
        self.mutation_temperature = mutation_temperature
        self.crossover_temperature = crossover_temperature

        self.kwargs = kwargs

        selections = ["best", "random", "softmax"]
        if selection not in selections:
            raise ValueError(f"Parameter `selection` should be between {selections}")
        self.selection = selection
        self.selection_temperature = selection_temperature

        self.mutation_programs = {}
        self.crossover_programs = {}

    def select_candidate(self, candidates):
        """Select a candidate from the list based on the selection strategy.

        Args:
            candidates (list): List of candidate dictionaries with 'reward' key

        Returns:
            The selected candidate, or None if the list is empty
        """
        if not candidates:
            return None

        if self.selection == "random":
            return random.choice(candidates)
        elif self.selection == "best":
            return sorted(
                candidates,
                key=lambda x: x.get("reward", 0),
                reverse=True,
            )[0]
        elif self.selection == "softmax":
            rewards = numpy.array(
                [candidate.get("reward", 0) for candidate in candidates]
            )
            scaled_rewards = rewards / self.selection_temperature
            exp_rewards = numpy.exp(scaled_rewards - numpy.max(scaled_rewards))
            probabilities = exp_rewards / numpy.sum(exp_rewards)
            return numpy.random.choice(
                candidates,
                size=1,
                replace=False,
                p=probabilities,
            ).tolist()[0]

    async def on_batch_begin(
        self,
        step,
        epoch,
        trainable_variables,
    ):
        """Called at the beginning of a batch.

        Implements selection strategies for choosing which candidate to evolve.

        Args:
            step (int): The batch number
            epoch (int): The epoch number
            trainable_variables (list): The list of trainable variables
        """
        for trainable_variable in trainable_variables:
            best_candidates = trainable_variable.get("best_candidates")
            if epoch == 0:
                seed_candidates = trainable_variable.get("seed_candidates")
                if len(seed_candidates) > 0:
                    seed_candidate = random.choice(seed_candidates)
                    trainable_variable.update(
                        {
                            **seed_candidate,
                        },
                    )
            else:
                if len(best_candidates) > 0:
                    best_candidate = self.select_candidate(best_candidates)
                    best_candidate = out_mask_json(
                        best_candidate,
                        mask=["reward"],
                    )
                    trainable_variable.update(
                        {
                            **best_candidate,
                        },
                    )
                else:
                    seed_candidates = trainable_variable.get("seed_candidates")
                    if len(seed_candidates) > 0:
                        seed_candidate = random.choice(seed_candidates)
                        trainable_variable.update(
                            {
                                **seed_candidate,
                            },
                        )
            trainable_variable.update(
                {
                    "nb_visit": 0,
                    "cumulative_reward": 0.0,
                },
            )

    async def propose_new_candidates(
        self,
        step,
        trainable_variables,
        x=None,
        y=None,
        y_pred=None,
        training=False,
    ):
        """Generate new candidates using mutation or crossover strategy.

        This method selects which variable to update, chooses between mutation
        and crossover based on merging_rate, and applies the appropriate operation.

        Args:
            step (int): The current training step
            trainable_variables (list): List of trainable variables to optimize
            x: Input data batch
            y: Ground truth data batch
            y_pred: Predicted outputs from the current model
            training (bool): Whether in training mode
        """
        variable_name_to_update = await self.select_variable_name_to_update(
            trainable_variables,
        )

        strategy = await self.select_evolving_strategy()

        for trainable_variable in trainable_variables:
            if trainable_variable.name == variable_name_to_update:
                best_candidates = trainable_variable.get("best_candidates")
                selected_candidate = self.select_candidate(best_candidates)

                if strategy == "mutation":
                    new_candidate = await self.mutate_candidate(
                        step,
                        trainable_variable,
                        selected_candidate,
                        x=x,
                        y=y,
                        y_pred=y_pred,
                        training=training,
                    )
                elif strategy == "crossover":
                    if len(best_candidates) >= 2:
                        # Select another candidate for crossover
                        other_candidate = await self.select_candidate_to_merge(
                            step,
                            trainable_variable,
                        )
                        new_candidate = await self.merge_candidate(
                            step,
                            trainable_variable,
                            selected_candidate,
                            other_candidate,
                            x=x,
                            y=y,
                            y_pred=y_pred,
                            training=training,
                        )
                    else:
                        # Fallback to mutation if not enough candidates to merge
                        new_candidate = await self.mutate_candidate(
                            step,
                            trainable_variable,
                            selected_candidate,
                            x=x,
                            y=y,
                            y_pred=y_pred,
                            training=training,
                        )

                await self.assign_candidate(
                    trainable_variable,
                    new_candidate=new_candidate,
                )

    async def mutate_candidate(
        self,
        step,
        trainable_variable,
        selected_candidate,
        x=None,
        y=None,
        y_pred=None,
        training=False,
    ):
        """Apply mutation to generate a new candidate.

        Subclasses must implement this method to define their mutation logic.

        Args:
            step (int): The current training step
            trainable_variable: The trainable variable (for metadata access)
            selected_candidate: The selected candidate to mutate
            x: Input data batch
            y: Ground truth data batch
            y_pred: Predicted outputs from the current model
            training (bool): Whether in training mode

        Returns:
            The mutated candidate
        """
        raise NotImplementedError(
            "EvolutionaryOptimizer subclasses must implement `mutate_candidate()`."
        )

    async def merge_candidate(
        self,
        step,
        trainable_variable,
        current_candidate,
        other_candidate,
        x=None,
        y=None,
        y_pred=None,
        training=False,
    ):
        """Apply crossover to merge two selected candidates.

        Subclasses must implement this method to define their crossover logic.

        Args:
            step (int): The current training step
            trainable_variable: The trainable variable (for metadata access)
            current_candidate: First selected candidate to merge
            other_candidate: Second selected candidate to merge
            x: Input data batch
            y: Ground truth data batch
            y_pred: Predicted outputs from the current model
            training (bool): Whether in training mode

        Returns:
            The merged candidate
        """
        raise NotImplementedError(
            "EvolutionaryOptimizer subclasses must implement `merge_candidate()`."
        )

    async def select_evolving_strategy(self):
        """Select between mutation and crossover based on merging_rate.

        Returns:
            str: Either "mutation" or "crossover"
        """
        rand = random.random()
        if rand > (self.merging_rate * self.epochs):
            return "mutation"
        else:
            return "crossover"

    def get_config(self):
        base_config = super().get_config()
        config = {
            "merging_rate": self.merging_rate,
            "mutation_temperature": self.mutation_temperature,
            "crossover_temperature": self.crossover_temperature,
            "selection": self.selection,
            "selection_temperature": self.selection_temperature,
        }
        language_model_config = {
            "language_model": serialization_lib.serialize_synalinks_object(
                self.language_model,
            )
        }
        return {**base_config, **config, **language_model_config}

    @classmethod
    def from_config(cls, config):
        language_model = serialization_lib.deserialize_synalinks_object(
            config.pop("language_model"),
        )
        return cls(language_model=language_model, **config)



# Modified from: keras/src/tree/tree_test.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import functools
from collections import OrderedDict
from collections import defaultdict
from collections import deque
from collections import namedtuple

import numpy as np
from absl.testing import parameterized

from synalinks.src import testing
from synalinks.src.tree.tree_api import MAP_TO_NONE
from synalinks.src.utils.module_utils import optree
from synalinks.src.utils.tracking import TrackedDict
from synalinks.src.utils.tracking import TrackedList
from synalinks.src.utils.tracking import TrackedSet

TEST_CASES = []

if optree.available:
    from synalinks.src.tree import optree_impl

    TEST_CASES += [
        {
            "testcase_name": "optree",
            "t": optree_impl,
        },
    ]

Empty = namedtuple("Empty", [])
Point = namedtuple("Point", ["x", "y"])
OtherPoint = namedtuple("OtherPoint", ["x", "y"])


def default_value():
    return None


class Visitor:
    def __init__(self, func):
        self.func = func
        self.visited_list = []

    def __call__(self, x):
        self.visited_list.append(x)
        return self.func(x)

    def visited(self):
        ret = self.visited_list
        self.visited_list = []
        return ret


@parameterized.named_parameters(TEST_CASES)
class TreeTest(testing.TestCase):
    # def setUp(self):
    #     if dmtree.available and optree.available:
    #         # If both are available, the annotation on the Keras tracking
    #         # wrappers will have used optree. For testing purposes, we need to
    #         # also register them with dm-tree.
    #         from synalinks.src.tree import dmtree_impl

    #         dmtree_impl.register_tree_node_class(TrackedList)
    #         dmtree_impl.register_tree_node_class(TrackedSet)
    #         dmtree_impl.register_tree_node_class(TrackedDict)
    #     super().setUp()

    def assertEqualStrict(self, a, b):
        self.assertEqual(a, b)
        self.assertEqual(type(a), type(b))
        if isinstance(a, OrderedDict):
            # Verify order.
            self.assertEqual(a.items(), b.items())
        elif isinstance(a, defaultdict):
            self.assertEqual(a.default_factory, b.default_factory)
        # Recurse
        if isinstance(a, (tuple, list, deque)):
            for sub_a, sub_b in zip(a, b):
                self.assertEqualStrict(sub_a, sub_b)
        elif isinstance(a, dict):
            for k in a:
                self.assertEqualStrict(a[k], b[k])

    def is_dmtree(self, tree_impl):
        # if dmtree.available:
        #     from synalinks.src.tree import dmtree_impl

        #     return tree_impl is dmtree_impl
        return False

    def test_is_nested(self, t):
        # Non-nested.
        self.assertFalse(t.is_nested(1))
        self.assertFalse(t.is_nested("1234"))
        self.assertFalse(t.is_nested(b"1234"))
        self.assertFalse(t.is_nested(bytearray("1234", "ascii")))
        self.assertFalse(t.is_nested(set([1, 2])))

        # Standard structures.
        self.assertTrue(t.is_nested(()))
        self.assertTrue(t.is_nested((1,)))
        self.assertTrue(t.is_nested((1, 2)))
        self.assertTrue(t.is_nested([]))
        self.assertTrue(t.is_nested([1]))
        self.assertTrue(t.is_nested([1, 2]))
        self.assertTrue(t.is_nested(deque([])))
        self.assertTrue(t.is_nested(deque([1])))
        self.assertTrue(t.is_nested(deque([1, 2])))
        self.assertTrue(t.is_nested(Empty()))
        self.assertTrue(t.is_nested(Point(x=1, y=2)))
        self.assertTrue(t.is_nested({}))
        self.assertTrue(t.is_nested({"a": 1}))
        self.assertTrue(t.is_nested({"b": 2, "a": 1}))
        self.assertTrue(t.is_nested(OrderedDict()))
        self.assertTrue(t.is_nested(OrderedDict([("a", 1)])))
        self.assertTrue(t.is_nested(OrderedDict([("b", 2), ("a", 1)])))
        self.assertTrue(t.is_nested(defaultdict(default_value)))
        self.assertTrue(t.is_nested(defaultdict(default_value, [("a", 1)])))
        self.assertTrue(t.is_nested(defaultdict(default_value, [("b", 2), ("a", 1)])))

        # Keras tracking wrappers.
        self.assertTrue(t.is_nested(TrackedList([])))
        self.assertTrue(t.is_nested(TrackedList([1])))
        self.assertTrue(t.is_nested(TrackedList([1, 2])))
        self.assertTrue(t.is_nested(TrackedSet([])))
        self.assertTrue(t.is_nested(TrackedSet([1])))
        self.assertTrue(t.is_nested(TrackedSet([1, 2])))
        self.assertTrue(t.is_nested(TrackedDict({})))
        self.assertTrue(t.is_nested(TrackedDict({"a": 1})))
        self.assertTrue(t.is_nested(TrackedDict({"b": 2, "a": 1})))

    def test_flatten(self, t):
        # Non-nested.
        self.assertEqualStrict(t.flatten(1), [1])

        # Standard structures.
        self.assertEqualStrict(t.flatten(()), [])
        self.assertEqualStrict(t.flatten((1,)), [1])
        self.assertEqualStrict(t.flatten((1, 2)), [1, 2])
        self.assertEqualStrict(t.flatten([]), [])
        self.assertEqualStrict(t.flatten([1]), [1])
        self.assertEqualStrict(t.flatten([1, 2]), [1, 2])
        self.assertEqualStrict(t.flatten(deque([])), [])
        self.assertEqualStrict(t.flatten(deque([1])), [1])
        self.assertEqualStrict(t.flatten(deque([1, 2])), [1, 2])
        self.assertEqualStrict(t.flatten(Empty()), [])
        self.assertEqualStrict(t.flatten(Point(y=2, x=1)), [1, 2])
        self.assertEqualStrict(t.flatten({}), [])
        self.assertEqualStrict(t.flatten({"a": 1}), [1])
        self.assertEqualStrict(t.flatten({"b": 2, "a": 1}), [1, 2])
        self.assertEqualStrict(
            t.flatten(OrderedDict()),
            [],
        )
        self.assertEqualStrict(
            t.flatten(OrderedDict([("a", 1)])),
            [1],
        )
        self.assertEqualStrict(
            t.flatten(OrderedDict([("b", 2), ("a", 1)])),
            [2, 1],
        )
        self.assertEqualStrict(
            t.flatten(defaultdict(default_value)),
            [],
        )
        self.assertEqualStrict(
            t.flatten(defaultdict(default_value, [("a", 1)])),
            [1],
        )
        self.assertEqualStrict(
            t.flatten(defaultdict(default_value, [("b", 2), ("a", 1)])),
            [1, 2],
        )

        # Keras tracking wrappers.
        self.assertEqualStrict(t.flatten(TrackedList([])), [])
        self.assertEqualStrict(t.flatten(TrackedList([1])), [1])
        self.assertEqualStrict(t.flatten(TrackedList([1, 2])), [1, 2])
        self.assertEqualStrict(t.flatten(TrackedSet([])), [])
        self.assertEqualStrict(t.flatten(TrackedSet([1])), [1])
        self.assertEqualStrict(sorted(t.flatten(TrackedSet([1, 2]))), [1, 2])
        self.assertEqualStrict(t.flatten(TrackedDict({})), [])
        self.assertEqualStrict(t.flatten(TrackedDict({"a": 1})), [1])
        self.assertEqualStrict(t.flatten(TrackedDict({"b": 2, "a": 1})), [1, 2])

        # Deeper nested structures.
        self.assertEqualStrict(
            t.flatten(
                (
                    {"b": [2, 3], "a": (1,)},
                    TrackedDict({"x": 4, "y": TrackedList([5, 6])}),
                    TrackedSet([7]),
                    Point(y=9, x=8),
                    np.array([10]),
                )
            ),
            [1, 2, 3, 4, 5, 6, 7, 8, 9, np.array([10])],
        )

    def test_map_structure_with_multiple_structures(self, t):
        def f2(x, y):
            return x + y if isinstance(x, int) and isinstance(y, int) else None

        # Non-nested.
        self.assertEqualStrict(t.map_structure(f2, 1, 10), 11)

        # Standard structures.
        self.assertEqualStrict(t.map_structure(f2, ()), ())
        self.assertEqualStrict(t.map_structure(f2, (1,), (10,)), (11,))
        self.assertEqualStrict(t.map_structure(f2, (1, 2), (10, 20)), (11, 22))
        self.assertEqualStrict(t.map_structure(f2, []), [])
        self.assertEqualStrict(t.map_structure(f2, [1], [10]), [11])
        self.assertEqualStrict(t.map_structure(f2, [1, 2], [10, 20]), [11, 22])
        self.assertEqualStrict(t.map_structure(f2, deque([])), deque([]))
        self.assertEqualStrict(t.map_structure(f2, deque([1]), deque([10])), deque([11]))
        self.assertEqualStrict(
            t.map_structure(f2, deque([1, 2]), deque([10, 20])), deque([11, 22])
        )
        self.assertEqualStrict(t.map_structure(f2, Empty()), Empty())
        self.assertEqualStrict(
            t.map_structure(f2, Point(y=2, x=1), Point(x=10, y=20)),
            Point(x=11, y=22),
        )
        self.assertEqualStrict(t.map_structure(f2, {}), {})
        self.assertEqualStrict(t.map_structure(f2, {"a": 1}, {"a": 10}), {"a": 11})
        self.assertEqualStrict(
            t.map_structure(f2, {"b": 2, "a": 1}, {"a": 10, "b": 20}),
            {"a": 11, "b": 22},
        )
        self.assertEqualStrict(
            t.map_structure(f2, OrderedDict()),
            OrderedDict(),
        )
        self.assertEqualStrict(
            t.map_structure(f2, OrderedDict([("a", 1)]), OrderedDict([("a", 10)])),
            OrderedDict([("a", 11)]),
        )
        self.assertEqualStrict(
            t.map_structure(
                f2,
                OrderedDict([("b", 2), ("a", 1)]),
                OrderedDict([("b", 20), ("a", 10)]),
            ),
            OrderedDict([("b", 22), ("a", 11)]),
        )
        self.assertEqualStrict(
            t.map_structure(f2, defaultdict(default_value), defaultdict(default_value)),
            defaultdict(default_value),
        )
        self.assertEqualStrict(
            t.map_structure(
                f2,
                defaultdict(default_value, [("a", 1)]),
                defaultdict(default_value, [("a", 10)]),
            ),
            defaultdict(default_value, [("a", 11)]),
        )
        self.assertEqualStrict(
            t.map_structure(
                f2,
                defaultdict(default_value, [("b", 2), ("a", 1)]),
                defaultdict(default_value, [("a", 10), ("b", 20)]),
            ),
            defaultdict(default_value, [("a", 11), ("b", 22)]),
        )

        # Keras tracking wrappers.
        self.assertEqualStrict(
            t.map_structure(
                f2,
                TrackedList([]),
                TrackedList([]),
            ),
            TrackedList([]),
        )
        self.assertEqualStrict(
            t.map_structure(
                f2,
                TrackedList([1]),
                TrackedList([10]),
            ),
            TrackedList([11]),
        )
        self.assertEqualStrict(
            t.map_structure(
                f2,
                TrackedList([1, 2]),
                TrackedList([10, 20]),
            ),
            TrackedList([11, 22]),
        )

        # Known limitation of the dm-tree implementation:
        # Registered classes are not handled when mapping multiple
        # structures at once. TrackedSet is the only problematic one.
        if not self.is_dmtree(t):
            self.assertEqualStrict(
                t.map_structure(
                    f2,
                    TrackedSet([]),
                    TrackedSet([]),
                ),
                TrackedSet([]),
            )
            self.assertEqualStrict(
                t.map_structure(
                    f2,
                    TrackedSet([1]),
                    TrackedSet([10]),
                ),
                TrackedSet([11]),
            )
            self.assertEqualStrict(
                t.map_structure(
                    f2,
                    TrackedSet([1, 2]),
                    TrackedSet([10, 20]),
                ),
                TrackedSet([11, 22]),
            )

        self.assertEqualStrict(
            t.map_structure(
                f2,
                TrackedDict({}),
                TrackedDict({}),
            ),
            TrackedDict({}),
        )
        self.assertEqualStrict(
            t.map_structure(
                f2,
                TrackedDict({"a": 1}),
                TrackedDict({"a": 10}),
            ),
            TrackedDict({"a": 11}),
        )
        self.assertEqualStrict(
            t.map_structure(
                f2,
                TrackedDict({"b": 2, "a": 1}),
                TrackedDict({"a": 10, "b": 20}),
            ),
            TrackedDict({"a": 11, "b": 22}),
        )

        # Deeper nested structures.
        self.assertEqualStrict(
            t.map_structure(
                f2,
                (
                    {"b": [2, 3], "a": (1,)},
                    TrackedDict({"x": 4, "y": TrackedList([5, 6])}),
                    TrackedSet([7]),
                    Point(y=9, x=8),
                    np.array([10]),
                ),
                (
                    {"b": [20, 30], "a": (10,)},
                    TrackedDict({"x": 40, "y": TrackedList([50, 60])}),
                    TrackedSet([70]),
                    Point(y=90, x=80),
                    np.array([100]),
                ),
            ),
            (
                {"b": [22, 33], "a": (11,)},
                TrackedDict({"x": 44, "y": TrackedList([55, 66])}),
                # Known limitation of the dm-tree implementation:
                # Registered classes are not handled when mapping multiple
                # structures at once. TrackedSet is the only problematic one.
                None if self.is_dmtree(t) else TrackedSet([77]),
                Point(y=99, x=88),
                None,
            ),
        )

        # Error cases.

        # list, tuple, deque and namedtuple are not considered equivalent.
        # Test all 6 combinations:
        # tuple, list.
        with self.assertRaisesRegex(ValueError, "tuple"):
            t.map_structure(f2, (), [])
        # tuple, deque.
        with self.assertRaisesRegex(ValueError, "tuple"):
            t.map_structure(f2, (), deque())
        # tuple, namedtuple.
        with self.assertRaisesRegex(ValueError, "tuple"):
            t.map_structure(f2, (), Empty())
        # list, deque.
        with self.assertRaisesRegex(ValueError, "list"):
            t.map_structure(f2, [], deque())
        # list, namedtuple.
        with self.assertRaisesRegex(ValueError, "list"):
            t.map_structure(f2, [], Empty())
        # deque, namedtuple.
        with self.assertRaisesRegex(ValueError, "deque"):
            t.map_structure(f2, deque(), Empty())

        # Equivalent namedtuples don't match.
        with self.assertRaisesRegex(ValueError, "namedtuple"):
            t.map_structure(f2, Point(x=1, y=2), OtherPoint(x=10, y=20))

        # Mismatched counts.
        with self.assertRaisesRegex(ValueError, "(number|[Aa]rity)"):
            t.map_structure(f2, (1, 2), (1,))
        with self.assertRaisesRegex(ValueError, "(number|[Aa]rity)"):
            t.map_structure(f2, [1, 2], [1])
        with self.assertRaisesRegex(ValueError, "(number|[Aa]rity)"):
            t.map_structure(f2, deque([1, 2]), deque([1]))

        # dict, OrderedDict, defaultdict are considered equivalent, but the
        # returned type is the first one. Test all 6 combinations (3 type
        # combinations plus the order).
        # dict, OrderedDict yields dict.
        self.assertEqualStrict(
            t.map_structure(f2, {"a": 1, "b": 2}, OrderedDict([("b", 20), ("a", 10)])),
            {"a": 11, "b": 22},
        )
        # OrderedDict, dict yields OrderedDict with same order.
        self.assertEqualStrict(
            t.map_structure(
                f2,
                OrderedDict([("b", 2), ("a", 1)]),
                {"a": 10, "b": 20},
            ),
            OrderedDict([("b", 22), ("a", 11)]),
        )
        # dict, defaultdict yields dict.
        self.assertEqualStrict(
            t.map_structure(
                f2,
                {"a": 1, "b": 2},
                defaultdict(default_value, [("b", 20), ("a", 10)]),
            ),
            {"a": 11, "b": 22},
        )
        # defaultdict, dict yields defaultdict.
        self.assertEqualStrict(
            t.map_structure(
                f2,
                defaultdict(default_value, [("b", 2), ("a", 1)]),
                {"a": 10, "b": 20},
            ),
            defaultdict(default_value, [("a", 11), ("b", 22)]),
        )
        # defaultdict, OrderedDict yields defaultdict.
        self.assertEqualStrict(
            t.map_structure(
                f2,
                defaultdict(default_value, [("a", 1), ("b", 2)]),
                OrderedDict([("b", 20), ("a", 10)]),
            ),
            defaultdict(default_value, [("a", 11), ("b", 22)]),
        )
        # OrderedDict, defaultdict yields OrderedDict with same order.
        self.assertEqualStrict(
            t.map_structure(
                f2,
                OrderedDict([("b", 2), ("a", 1)]),
                defaultdict(default_value, [("a", 10), ("b", 20)]),
            ),
            OrderedDict([("b", 22), ("a", 11)]),
        )

        # Multiple OrderedDicts with same keys but different orders, the order
        # of the first one prevails.
        self.assertEqualStrict(
            t.map_structure(
                f2,
                OrderedDict([("b", 2), ("a", 1)]),
                OrderedDict([("a", 10), ("b", 20)]),
            ),
            OrderedDict([("b", 22), ("a", 11)]),
        )

        # Mismatched keys
        with self.assertRaisesRegex(ValueError, "key"):
            t.map_structure(f2, {"a": 1, "b": 2}, {"a": 1})
        with self.assertRaisesRegex(ValueError, "key"):
            t.map_structure(
                f2,
                defaultdict(default_value, [("a", 1), ("b", 2)]),
                defaultdict(default_value, [("a", 10)]),
            )
        with self.assertRaisesRegex(ValueError, "key"):
            t.map_structure(
                f2, OrderedDict([("a", 1), ("b", 2)]), OrderedDict([("a", 10)])
            )

    def test_assert_same_structure(self, t):
        # Non-nested.
        t.assert_same_structure(1, 10)

        # Standard structures.
        t.assert_same_structure((), ())
        t.assert_same_structure((1,), (10,))
        t.assert_same_structure((1, 2), (10, 20))
        t.assert_same_structure([], [])
        t.assert_same_structure([1], [10])
        t.assert_same_structure([1, 2], [10, 20])
        t.assert_same_structure(deque([]), deque([]))
        t.assert_same_structure(deque([1]), deque([1]))
        t.assert_same_structure(deque([1, 2]), deque([10, 20]))
        t.assert_same_structure(Empty(), Empty())
        t.assert_same_structure(Point(y=1, x=2), Point(x=10, y=20))
        t.assert_same_structure({}, {})
        t.assert_same_structure({"a": 1}, {"a": 10})
        t.assert_same_structure({"b": 2, "a": 1}, {"a": 10, "b": 20})
        t.assert_same_structure(OrderedDict(), OrderedDict())
        t.assert_same_structure(OrderedDict([("a", 1)]), OrderedDict([("a", 10)]))
        t.assert_same_structure(
            OrderedDict([("b", 1), ("a", 2)]),
            OrderedDict([("b", 10), ("a", 20)]),
        )
        t.assert_same_structure(defaultdict(default_value), defaultdict(default_value))
        t.assert_same_paths(
            defaultdict(default_value, [("a", 1)]),
            defaultdict(default_value, [("a", 10)]),
        )
        t.assert_same_paths(
            defaultdict(default_value, [("b", 1), ("a", 2)]),
            defaultdict(default_value, [("a", 10), ("b", 20)]),
        )

        # Keras tracking wrappers.
        t.assert_same_structure(
            TrackedList([]),
            TrackedList([]),
        )
        t.assert_same_structure(
            TrackedList([1]),
            TrackedList([10]),
        )
        t.assert_same_structure(
            TrackedList([1, 2]),
            TrackedList([10, 20]),
        )
        t.assert_same_structure(
            TrackedSet([]),
            TrackedSet([]),
        )
        t.assert_same_structure(
            TrackedSet([1]),
            TrackedSet([10]),
        )
        t.assert_same_structure(
            TrackedSet([1, 2]),
            TrackedSet([10, 20]),
        )
        t.assert_same_structure(
            TrackedDict({}),
            TrackedDict({}),
        )
        t.assert_same_structure(
            TrackedDict({"a": 1}),
            TrackedDict({"a": 10}),
        )
        t.assert_same_structure(
            TrackedDict({"b": 2, "a": 1}),
            TrackedDict({"a": 10, "b": 20}),
        )

        # Deeper nested structures.
        t.assert_same_structure(
            (
                {"b": [2, 3], "a": (1,)},
                TrackedDict({"x": 4, "y": TrackedList([5, 6])}),
                TrackedSet([7]),
                Point(y=9, x=8),
                np.array([10]),
            ),
            (
                {"b": [20, 30], "a": (10,)},
                TrackedDict({"x": 40, "y": TrackedList([50, 60])}),
                TrackedSet([70]),
                Point(y=90, x=80),
                np.array([100]),
            ),
        )

        # Error cases.

        # Non-nested vs. nested.
        with self.assertRaisesRegex(ValueError, "don't have the same nested"):
            t.assert_same_structure(1, ())
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*tuple"):
            t.assert_same_structure((), 1)
        with self.assertRaisesRegex(ValueError, "don't have the same nested"):
            t.assert_same_structure(1, [])
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*list"):
            t.assert_same_structure([], 1)
        with self.assertRaisesRegex(ValueError, "don't have the same nested"):
            t.assert_same_structure(1, deque([]))
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*deque"):
            t.assert_same_structure(deque([]), 1)
        with self.assertRaisesRegex(ValueError, "don't have the same nested"):
            t.assert_same_structure(1, Empty())
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*(Empty|tuple)"):
            t.assert_same_structure(Empty(), 1)
        with self.assertRaisesRegex(ValueError, "don't have the same nested"):
            t.assert_same_structure(1, Point(x=1, y=2))
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*(Point|tuple)"):
            t.assert_same_structure(Point(x=1, y=2), 1)
        with self.assertRaisesRegex(ValueError, "don't have the same nested"):
            t.assert_same_structure(1, {})
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*dict"):
            t.assert_same_structure({}, 1)
        with self.assertRaisesRegex(ValueError, "don't have the same nested"):
            t.assert_same_structure(1, OrderedDict())
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*OrderedDict"):
            t.assert_same_structure(OrderedDict(), 1)
        with self.assertRaisesRegex(ValueError, "don't have the same nested"):
            t.assert_same_structure(1, defaultdict(default_value))
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*defaultdict"):
            t.assert_same_structure(defaultdict(default_value), 1)

        # Non-nested vs. Keras tracking wrappers.
        with self.assertRaisesRegex(ValueError, "(nested|TrackedList)"):
            t.assert_same_structure(1, TrackedList([]))
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*TrackedList"):
            t.assert_same_structure(TrackedList([]), 1)
        with self.assertRaisesRegex(ValueError, "(nested|TrackedSet)"):
            t.assert_same_structure(1, TrackedSet([]))
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*TrackedSet"):
            t.assert_same_structure(TrackedSet([]), 1)
        with self.assertRaisesRegex(ValueError, "(nested|TrackedDict)"):
            t.assert_same_structure(1, TrackedDict([]))
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*TrackedDict"):
            t.assert_same_structure(TrackedDict([]), 1)

        # list, tuple, deque and namedtuple are not considered equivalent.
        # Test all 6 combinations:
        # tuple, list.
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*tuple"):
            t.assert_same_structure((), [])
        # tuple, deque.
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*tuple"):
            t.assert_same_structure((), deque())
        # tuple, namedtuple.
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*tuple"):
            t.assert_same_structure((), Empty())
        # list, deque.
        with self.assertRaisesRegex(ValueError, "list"):
            t.assert_same_structure([], deque())
        # list, namedtuple.
        with self.assertRaisesRegex(ValueError, "list"):
            t.assert_same_structure([], Empty())
        # deque, namedtuple.
        with self.assertRaisesRegex(ValueError, "deque"):
            t.assert_same_structure(deque(), Empty())

        # Equivalent namedtuples don't match.
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*[. ]Point"):
            t.assert_same_structure(Point(x=1, y=2), OtherPoint(x=10, y=20))

        # Mismatched counts.
        with self.assertRaisesRegex(ValueError, "[Aa]rity mismatch"):
            t.assert_same_structure((1, 2), (1,))
        with self.assertRaisesRegex(ValueError, "[Aa]rity mismatch"):
            t.assert_same_structure([1, 2], [1])
        with self.assertRaisesRegex(ValueError, "[Aa]rity mismatch"):
            t.assert_same_structure(deque([1, 2]), deque([1]))

        # Mismatched counts with Keras tracking wrappers.
        with self.assertRaisesRegex(ValueError, "[Aa]rity mismatch"):
            t.assert_same_structure(TrackedList([1, 2]), TrackedList([1]))
        with self.assertRaisesRegex(ValueError, "[Aa]rity mismatch"):
            t.assert_same_structure(TrackedSet([1, 2]), TrackedSet([1]))

        # dict, OrderedDict, defaultdict are considered equivalent.
        # Test all 6 combinations (3 type combinations plus the order).
        # dict, OrderedDict.
        t.assert_same_structure({"a": 1, "b": 2}, OrderedDict([("b", 20), ("a", 10)]))
        # OrderedDict, dict.
        t.assert_same_structure(OrderedDict([("b", 20), ("a", 10)]), {"a": 1, "b": 2})
        # dict, defaultdict.
        t.assert_same_structure(
            {"a": 1, "b": 2},
            defaultdict(default_value, [("b", 20), ("a", 10)]),
        )
        # defaultdict, dict.
        t.assert_same_structure(
            defaultdict(default_value, [("b", 20), ("a", 10)]),
            {"a": 1, "b": 2},
        )
        # defaultdict, OrderedDict.
        t.assert_same_structure(
            defaultdict(default_value, [("a", 1), ("b", 2)]),
            OrderedDict([("b", 20), ("a", 10)]),
        )
        # OrderedDict, defaultdict.
        t.assert_same_structure(
            OrderedDict([("b", 2), ("a", 1)]),
            defaultdict(default_value, [("a", 10), ("b", 20)]),
        )

        # Two OrderedDicts with same keys but different orders.
        t.assert_same_structure(
            OrderedDict([("b", 2), ("a", 1)]),
            OrderedDict([("a", 10), ("b", 20)]),
        )

        # Keras tracking wrappers are not equivalent to the raw structures.
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*TrackedList"):
            t.assert_same_structure(TrackedList([1, 2]), list([10, 20]))
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*list"):
            t.assert_same_structure(list([1, 2]), TrackedList([10, 20]))
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*TrackedSet"):
            t.assert_same_structure(TrackedSet([1, 2]), list([10, 20]))
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*list"):
            t.assert_same_structure(list([1, 2]), TrackedSet([10, 20]))
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*TrackedDict"):
            t.assert_same_structure(TrackedDict({"b": 2, "a": 1}), {"a": 10, "b": 20})
        with self.assertRaisesRegex(ValueError, "[Ee]xpected.*dict"):
            t.assert_same_structure({"b": 2, "a": 1}, TrackedDict({"a": 10, "b": 20}))

        # Mismatched key count.
        with self.assertRaisesRegex(ValueError, "[Dd]ictionary key mismatch"):
            t.assert_same_structure(
                {"a": 1, "b": 2},
                {"a": 1},
            )
        with self.assertRaisesRegex(ValueError, "[Dd]ictionary key mismatch"):
            t.assert_same_structure(
                defaultdict(default_value, [("a", 1), ("b", 2)]),
                defaultdict(default_value, [("a", 10)]),
            )
        with self.assertRaisesRegex(ValueError, "[Dd]ictionary key mismatch"):
            t.assert_same_structure(
                OrderedDict([("a", 1), ("b", 2)]),
                OrderedDict([("a", 10)]),
            )

        # Mismatched keys.
        with self.assertRaisesRegex(ValueError, "[Dd]ictionary key mismatch"):
            t.assert_same_structure(
                {"a": 1},
                {"b": 2},
            )
        with self.assertRaisesRegex(ValueError, "[Dd]ictionary key mismatch"):
            t.assert_same_structure(
                defaultdict(default_value, [("a", 1)]),
                defaultdict(default_value, [("b", 2)]),
            )
        with self.assertRaisesRegex(ValueError, "[Dd]ictionary key mismatch"):
            t.assert_same_structure(
                OrderedDict([("a", 1)]),
                OrderedDict([("b", 2)]),
            )

        # Mismatched key count and keys with TrackedDict.
        with self.assertRaisesRegex(ValueError, "Mismatch custom node data"):
            t.assert_same_structure(
                TrackedDict({"a": 1, "b": 2}),
                TrackedDict({"a": 1}),
            )
        with self.assertRaisesRegex(ValueError, "Mismatch custom node data"):
            t.assert_same_structure(
                TrackedDict({"a": 1}),
                TrackedDict({"b": 2}),
            )

    def test_assert_same_paths(self, t):
        # Non-nested.
        t.assert_same_paths(1, 10)

        # Standard structures.
        t.assert_same_paths((), ())
        t.assert_same_paths((1,), (10,))
        t.assert_same_paths((1, 2), (10, 20))
        t.assert_same_paths([], [])
        t.assert_same_paths([1], [10])
        t.assert_same_paths([1, 2], [10, 20])
        t.assert_same_paths(deque([]), deque([]))
        t.assert_same_paths(deque([1]), deque([10]))
        t.assert_same_paths(deque([1, 2]), deque([10, 20]))
        t.assert_same_paths(Empty(), Empty())
        t.assert_same_paths(Point(y=2, x=1), Point(x=10, y=20))
        t.assert_same_paths({}, {})
        t.assert_same_paths({"a": 1}, {"a": 10})
        t.assert_same_paths({"b": None, "a": None}, {"a": 10, "b": 20})
        t.assert_same_paths(OrderedDict(), OrderedDict())
        t.assert_same_paths(OrderedDict([("a", 1)]), OrderedDict([("a", 10)]))
        t.assert_same_paths(
            OrderedDict([("b", 2), ("a", 1)]),
            OrderedDict([("a", 10), ("b", 20)]),
        )
        t.assert_same_paths(defaultdict(default_value), defaultdict(default_value))
        t.assert_same_paths(
            defaultdict(default_value, [("a", 1)]),
            defaultdict(default_value, [("a", 10)]),
        )
        t.assert_same_paths(
            defaultdict(default_value, [("b", 2), ("a", 1)]),
            defaultdict(default_value, [("a", 1), ("b", 2)]),
        )

        # Keras tracking wrappers.
        t.assert_same_paths(
            TrackedList([]),
            TrackedList([]),
        )
        t.assert_same_paths(
            TrackedList([1]),
            TrackedList([10]),
        )
        t.assert_same_paths(
            TrackedList([1, 2]),
            TrackedList([10, 20]),
        )
        t.assert_same_paths(
            TrackedSet([]),
            TrackedSet([]),
        )
        t.assert_same_paths(
            TrackedSet([1]),
            TrackedSet([10]),
        )
        t.assert_same_paths(
            TrackedSet([1, 2]),
            TrackedSet([10, 20]),
        )
        t.assert_same_paths(
            TrackedDict({}),
            TrackedDict({}),
        )
        t.assert_same_paths(
            TrackedDict({"a": 1}),
            TrackedDict({"a": 10}),
        )
        t.assert_same_paths(
            TrackedDict({"b": 2, "a": 1}),
            TrackedDict({"a": 10, "b": 20}),
        )

        # Deeper nested structures.
        t.assert_same_paths(
            (
                {"b": [2, 3], "a": (1,)},
                TrackedDict({"x": 4, "y": TrackedList([5, 6])}),
                TrackedSet([7]),
                Point(y=9, x=8),
                np.array([10]),
            ),
            (
                {"b": [20, 30], "a": (10,)},
                TrackedDict({"x": 40, "y": TrackedList([50, 60])}),
                TrackedSet([70]),
                Point(y=90, x=80),
                np.array([100]),
            ),
        )

        # list, tuple, deque and namedtuple have the same paths.
        # Test all 6 combinations:
        # tuple, list.
        t.assert_same_paths((), [])
        t.assert_same_paths([1, 2], (10, 20))
        # tuple, deque.
        t.assert_same_paths((), deque())
        t.assert_same_paths(deque([1, 2]), (10, 20))
        # tuple, namedtuple.
        t.assert_same_paths((), Empty())
        t.assert_same_paths(Point(x=1, y=2), (10, 20))
        # list, deque.
        t.assert_same_paths([], deque())
        t.assert_same_paths(deque([1, 2]), [10, 20])
        # list, namedtuple.
        t.assert_same_paths([], Empty())
        t.assert_same_paths(Point(x=None, y=20), [1, 2])
        # deque, namedtuple.
        t.assert_same_paths(deque(), Empty())
        t.assert_same_paths(Point(x=None, y=20), deque([1, 2]))

        # Equivalent namedtuples.
        t.assert_same_paths(Point(x=1, y=2), OtherPoint(x=None, y=20))

        # Mismatched counts.
        with self.assertRaisesRegex(ValueError, "don't have the same paths"):
            t.assert_same_paths((1, 2), (1,))
        with self.assertRaisesRegex(ValueError, "don't have the same paths"):
            t.assert_same_paths([1, 2], [1])
        with self.assertRaisesRegex(ValueError, "don't have the same paths"):
            t.assert_same_paths(deque([1, 2]), deque([1]))

        # dict, OrderedDict, defaultdict are considered equivalent. Test all 6
        # combinations (3 type combinations plus the order).
        # dict, OrderedDict.
        t.assert_same_paths({"a": 1, "b": 2}, OrderedDict([("b", 20), ("a", 10)]))
        # OrderedDict, dict.
        t.assert_same_paths(OrderedDict([("b", 20), ("a", 10)]), {"a": 1, "b": 2})
        # dict, defaultdict.
        t.assert_same_paths(
            {"a": 1, "b": 2},
            defaultdict(default_value, [("b", 20), ("a", 10)]),
        )
        # defaultdict, dict.
        t.assert_same_paths(
            defaultdict(default_value, [("b", 20), ("a", 10)]),
            {"a": 1, "b": 2},
        )
        # defaultdict, OrderedDict.
        t.assert_same_paths(
            defaultdict(default_value, [("a", 1), ("b", 2)]),
            OrderedDict([("b", 20), ("a", 10)]),
        )
        # OrderedDict, defaultdict.
        t.assert_same_paths(
            OrderedDict([("b", 2), ("a", 1)]),
            defaultdict(default_value, [("a", 10), ("b", 20)]),
        )

        # Two OrderedDicts with same keys but different orders.
        t.assert_same_paths(
            OrderedDict([("b", 2), ("a", 1)]),
            OrderedDict([("a", 10), ("b", 20)]),
        )

        # Keras tracking wrappers are equivalent to the raw structures.
        t.assert_same_paths(TrackedList([1, 2]), list([10, 20]))
        t.assert_same_paths(list([1, 2]), TrackedList([10, 20]))
        t.assert_same_paths(TrackedSet([1, 2]), list([10, 20]))
        t.assert_same_paths(list([1, 2]), TrackedSet([10, 20]))
        t.assert_same_paths(TrackedDict({"b": 2, "a": 1}), {"a": 10, "b": 20})
        t.assert_same_paths({"b": 2, "a": 1}, TrackedDict({"a": 10, "b": 20}))

        # Mismatched keys
        with self.assertRaisesRegex(ValueError, "don't have the same paths"):
            t.assert_same_paths({"a": 1, "b": 2}, {"a": 1})
        with self.assertRaisesRegex(ValueError, "don't have the same paths"):
            t.assert_same_paths(
                defaultdict(default_value, [("a", 1), ("b", 2)]),
                defaultdict(default_value, [("a", 10)]),
            )
        with self.assertRaisesRegex(ValueError, "don't have the same paths"):
            t.assert_same_paths(
                OrderedDict([("a", 1), ("b", 2)]), OrderedDict([("a", 10)])
            )

    def test_traverse_top_down(self, t):
        v = Visitor(lambda x: None if t.is_nested(x) else x + 10)

        # Non-nested.
        self.assertEqualStrict(t.traverse(v, 1), 11)
        self.assertEqualStrict(v.visited(), [1])

        # Standard structures.
        self.assertEqualStrict(t.traverse(v, ()), ())
        self.assertEqualStrict(v.visited(), [()])

        self.assertEqualStrict(t.traverse(v, (1,)), (11,))
        self.assertEqualStrict(v.visited(), [(1,), 1])

        self.assertEqualStrict(t.traverse(v, (1, 2)), (11, 12))
        self.assertEqualStrict(v.visited(), [(1, 2), 1, 2])

        self.assertEqualStrict(t.traverse(v, []), [])
        self.assertEqualStrict(v.visited(), [[]])

        self.assertEqualStrict(t.traverse(v, [1]), [11])
        self.assertEqualStrict(v.visited(), [[1], 1])

        self.assertEqualStrict(t.traverse(v, [1, 2]), [11, 12])
        self.assertEqualStrict(v.visited(), [[1, 2], 1, 2])

        self.assertEqualStrict(t.traverse(v, deque([])), deque([]))
        self.assertEqualStrict(v.visited(), [deque([])])

        self.assertEqualStrict(t.traverse(v, deque([1])), deque([11]))
        self.assertEqualStrict(v.visited(), [deque([1]), 1])

        self.assertEqualStrict(t.traverse(v, deque([1, 2])), deque([11, 12]))
        self.assertEqualStrict(v.visited(), [deque([1, 2]), 1, 2])

        self.assertEqualStrict(t.traverse(v, Empty()), Empty())
        self.assertEqualStrict(v.visited(), [Empty()])

        self.assertEqualStrict(t.traverse(v, Point(y=2, x=1)), Point(x=11, y=12))
        self.assertEqualStrict(v.visited(), [Point(x=1, y=2), 1, 2])

        self.assertEqualStrict(t.traverse(v, {}), {})
        self.assertEqualStrict(v.visited(), [{}])

        self.assertEqualStrict(t.traverse(v, {"a": 1}), {"a": 11})
        self.assertEqualStrict(v.visited(), [{"a": 1}, 1])

        self.assertEqualStrict(t.traverse(v, {"b": 2, "a": 1}), {"a": 11, "b": 12})
        self.assertEqualStrict(v.visited(), [{"a": 1, "b": 2}, 1, 2])

        self.assertEqualStrict(t.traverse(v, OrderedDict()), OrderedDict())
        self.assertEqualStrict(v.visited(), [OrderedDict()])

        self.assertEqualStrict(
            t.traverse(v, OrderedDict([("a", 1)])), OrderedDict([("a", 11)])
        )
        self.assertEqualStrict(v.visited(), [OrderedDict([("a", 1)]), 1])

        self.assertEqualStrict(
            t.traverse(v, OrderedDict([("b", 2), ("a", 1)])),
            OrderedDict([("b", 12), ("a", 11)]),
        )
        self.assertEqualStrict(v.visited(), [OrderedDict([("b", 2), ("a", 1)]), 2, 1])

        self.assertEqualStrict(
            t.traverse(v, defaultdict(default_value)),
            defaultdict(default_value),
        )
        self.assertEqualStrict(v.visited(), [defaultdict(default_value)])

        self.assertEqualStrict(
            t.traverse(v, defaultdict(default_value, [("a", 1)])),
            defaultdict(default_value, [("a", 11)]),
        )
        self.assertEqualStrict(v.visited(), [defaultdict(default_value, [("a", 1)]), 1])

        self.assertEqualStrict(
            t.traverse(v, defaultdict(default_value, [("b", 2), ("a", 1)])),
            defaultdict(default_value, [("a", 11), ("b", 12)]),
        )
        self.assertEqualStrict(
            v.visited(),
            [defaultdict(default_value, [("a", 1), ("b", 2)]), 1, 2],
        )

        # Keras tracking wrappers.
        self.assertEqualStrict(t.traverse(v, TrackedList([])), TrackedList([]))
        self.assertEqualStrict(v.visited(), [TrackedList([])])

        self.assertEqualStrict(t.traverse(v, TrackedList([1])), TrackedList([11]))
        self.assertEqualStrict(v.visited(), [TrackedList([1]), 1])

        self.assertEqualStrict(t.traverse(v, TrackedList([1, 2])), TrackedList([11, 12]))
        self.assertEqualStrict(v.visited(), [TrackedList([1, 2]), 1, 2])

        self.assertEqualStrict(t.traverse(v, TrackedSet([])), TrackedSet([]))
        self.assertEqualStrict(v.visited(), [TrackedSet([])])

        self.assertEqualStrict(t.traverse(v, TrackedSet([1])), TrackedSet([11]))
        self.assertEqualStrict(v.visited(), [TrackedSet([1]), 1])

        self.assertEqualStrict(t.traverse(v, TrackedSet([1, 2])), TrackedSet([11, 12]))
        visited = v.visited()
        self.assertEqualStrict(visited[0], TrackedSet([1, 2]))
        self.assertEqualStrict(sorted(visited[1:]), [1, 2])

        self.assertEqualStrict(
            t.traverse(v, TrackedDict()),
            TrackedDict(),
        )
        self.assertEqualStrict(v.visited(), [TrackedDict()])

        self.assertEqualStrict(
            t.traverse(v, TrackedDict({"a": 1})),
            TrackedDict({"a": 11}),
        )
        self.assertEqualStrict(v.visited(), [TrackedDict({"a": 1}), 1])

        self.assertEqualStrict(
            t.traverse(v, TrackedDict({"b": 2, "a": 1})),
            TrackedDict({"a": 11, "b": 12}),
        )
        self.assertEqualStrict(v.visited(), [TrackedDict({"a": 1, "b": 2}), 1, 2])

        # Deeper nested structures.
        self.assertEqualStrict(
            t.traverse(
                v,
                (
                    {"b": [2, 3], "a": (1,)},
                    TrackedDict({"x": 4, "y": TrackedList([5, 6])}),
                    TrackedSet([7]),
                    Point(y=9, x=8),
                    np.array([10]),
                ),
            ),
            (
                {"b": [12, 13], "a": (11,)},
                TrackedDict({"x": 14, "y": TrackedList([15, 16])}),
                TrackedSet([17]),
                Point(y=19, x=18),
                np.array([20]),
            ),
        )
        self.assertEqualStrict(
            v.visited(),
            [
                (
                    {"b": [2, 3], "a": (1,)},
                    TrackedDict({"x": 4, "y": TrackedList([5, 6])}),
                    TrackedSet([7]),
                    Point(y=9, x=8),
                    np.array([10]),
                ),
                {"b": [2, 3], "a": (1,)},
                (1,),
                1,
                [2, 3],
                2,
                3,
                TrackedDict({"x": 4, "y": TrackedList([5, 6])}),
                4,
                TrackedList([5, 6]),
                5,
                6,
                TrackedSet([7]),
                7,
                Point(x=8, y=9),
                8,
                9,
                np.array([10]),
            ],
        )

        # Error cases.
        with self.assertRaisesRegex(TypeError, "callable"):
            t.traverse("bad", [1, 2])

        # Children are not explored if structure is replaced with a leaf.
        v = Visitor(lambda x: "X" if isinstance(x, tuple) else None)
        self.assertEqualStrict(
            t.traverse(v, [(1, [2]), [3, (4, 5, 6)]]),
            ["X", [3, "X"]],
        )
        self.assertEqualStrict(
            v.visited(),
            [
                [(1, [2]), [3, (4, 5, 6)]],
                (1, [2]),
                [3, (4, 5, 6)],
                3,
                (4, 5, 6),
            ],
        )

        # Children are not explored if structure is replaced with structure.
        v = Visitor(lambda x: ("a", "b") if isinstance(x, tuple) else None)
        self.assertEqualStrict(
            t.traverse(v, [(1, [2]), [3, (4, 5, 6)]]),
            [("a", "b"), [3, ("a", "b")]],
        )
        self.assertEqualStrict(
            v.visited(),
            [
                [(1, [2]), [3, (4, 5, 6)]],
                (1, [2]),
                [3, (4, 5, 6)],
                3,
                (4, 5, 6),
            ],
        )

        # MAP_TO_NONE.
        v = Visitor(lambda x: MAP_TO_NONE if isinstance(x, tuple) else None)
        self.assertEqualStrict(
            t.traverse(v, [(1, [2]), [3, (4, 5, 6)]]),
            [None, [3, None]],
        )
        self.assertEqualStrict(
            v.visited(),
            [
                [(1, [2]), [3, (4, 5, 6)]],
                (1, [2]),
                [3, (4, 5, 6)],
                3,
                (4, 5, 6),
            ],
        )

    def test_traverse_bottom_up(self, t):
        v = Visitor(lambda x: None if t.is_nested(x) else x + 10)
        traverse_u = functools.partial(t.traverse, top_down=False)

        # Non-nested.
        self.assertEqualStrict(traverse_u(v, 1), 11)
        self.assertEqualStrict(v.visited(), [1])

        # Standard structures.
        self.assertEqualStrict(traverse_u(v, ()), ())
        self.assertEqualStrict(v.visited(), [()])

        self.assertEqualStrict(traverse_u(v, (1,)), (11,))
        self.assertEqualStrict(v.visited(), [1, (11,)])

        self.assertEqualStrict(traverse_u(v, (1, 2)), (11, 12))
        self.assertEqualStrict(v.visited(), [1, 2, (11, 12)])

        self.assertEqualStrict(traverse_u(v, []), [])
        self.assertEqualStrict(v.visited(), [[]])

        self.assertEqualStrict(traverse_u(v, [1]), [11])
        self.assertEqualStrict(v.visited(), [1, [11]])

        self.assertEqualStrict(traverse_u(v, [1, 2]), [11, 12])
        self.assertEqualStrict(v.visited(), [1, 2, [11, 12]])

        self.assertEqualStrict(traverse_u(v, deque([])), deque([]))
        self.assertEqualStrict(v.visited(), [deque([])])

        self.assertEqualStrict(traverse_u(v, deque([1])), deque([11]))
        self.assertEqualStrict(v.visited(), [1, deque([11])])

        self.assertEqualStrict(traverse_u(v, deque([1, 2])), deque([11, 12]))
        self.assertEqualStrict(v.visited(), [1, 2, deque([11, 12])])

        self.assertEqualStrict(traverse_u(v, Empty()), Empty())
        self.assertEqualStrict(v.visited(), [Empty()])

        self.assertEqualStrict(traverse_u(v, Point(y=2, x=1)), Point(x=11, y=12))
        self.assertEqualStrict(v.visited(), [1, 2, Point(x=11, y=12)])

        self.assertEqualStrict(traverse_u(v, {}), {})
        self.assertEqualStrict(v.visited(), [{}])

        self.assertEqualStrict(traverse_u(v, {"a": 1}), {"a": 11})
        self.assertEqualStrict(v.visited(), [1, {"a": 11}])

        self.assertEqualStrict(traverse_u(v, {"b": 2, "a": 1}), {"a": 11, "b": 12})
        self.assertEqualStrict(v.visited(), [1, 2, {"a": 11, "b": 12}])

        self.assertEqualStrict(traverse_u(v, OrderedDict()), OrderedDict())
        self.assertEqualStrict(v.visited(), [OrderedDict()])

        self.assertEqualStrict(
            traverse_u(v, OrderedDict([("a", 1)])), OrderedDict([("a", 11)])
        )
        self.assertEqualStrict(v.visited(), [1, OrderedDict([("a", 11)])])

        self.assertEqualStrict(
            traverse_u(v, OrderedDict([("b", 2), ("a", 1)])),
            OrderedDict([("b", 12), ("a", 11)]),
        )
        self.assertEqualStrict(v.visited(), [2, 1, OrderedDict([("b", 12), ("a", 11)])])

        self.assertEqualStrict(
            traverse_u(v, defaultdict(default_value)),
            defaultdict(default_value),
        )
        self.assertEqualStrict(v.visited(), [defaultdict(default_value)])

        self.assertEqualStrict(
            traverse_u(v, defaultdict(default_value, [("a", 1)])),
            defaultdict(default_value, [("a", 11)]),
        )
        self.assertEqualStrict(v.visited(), [1, defaultdict(default_value, [("a", 11)])])

        self.assertEqualStrict(
            traverse_u(v, defaultdict(default_value, [("b", 2), ("a", 1)])),
            defaultdict(default_value, [("a", 11), ("b", 12)]),
        )
        self.assertEqualStrict(
            v.visited(),
            [1, 2, defaultdict(default_value, [("a", 11), ("b", 12)])],
        )

        # Keras tracking wrappers.
        self.assertEqualStrict(traverse_u(v, TrackedList([])), TrackedList([]))
        self.assertEqualStrict(v.visited(), [TrackedList([])])

        self.assertEqualStrict(traverse_u(v, TrackedList([1])), TrackedList([11]))
        self.assertEqualStrict(v.visited(), [1, TrackedList([11])])

        self.assertEqualStrict(traverse_u(v, TrackedList([1, 2])), TrackedList([11, 12]))
        self.assertEqualStrict(v.visited(), [1, 2, TrackedList([11, 12])])

        self.assertEqualStrict(traverse_u(v, TrackedSet([])), TrackedSet([]))
        self.assertEqualStrict(v.visited(), [TrackedSet([])])

        self.assertEqualStrict(traverse_u(v, TrackedSet([1])), TrackedSet([11]))
        self.assertEqualStrict(v.visited(), [1, TrackedSet([11])])

        self.assertEqualStrict(traverse_u(v, TrackedSet([1, 2])), TrackedSet([11, 12]))
        visited = v.visited()
        self.assertEqualStrict(visited[-1], TrackedSet([11, 12]))
        self.assertEqualStrict(sorted(visited[:-1]), [1, 2])

        self.assertEqualStrict(
            traverse_u(v, TrackedDict()),
            TrackedDict(),
        )
        self.assertEqualStrict(v.visited(), [TrackedDict()])

        self.assertEqualStrict(
            traverse_u(v, TrackedDict({"a": 1})),
            TrackedDict({"a": 11}),
        )
        self.assertEqualStrict(v.visited(), [1, TrackedDict({"a": 11})])

        self.assertEqualStrict(
            traverse_u(v, TrackedDict({"b": 2, "a": 1})),
            TrackedDict({"a": 11, "b": 12}),
        )
        self.assertEqualStrict(v.visited(), [1, 2, TrackedDict({"a": 11, "b": 12})])

        # Deeper nested structures.
        self.assertEqualStrict(
            traverse_u(
                v,
                (
                    {"b": [2, 3], "a": (1,)},
                    TrackedDict({"x": 4, "y": TrackedList([5, 6])}),
                    TrackedSet([7]),
                    Point(y=9, x=8),
                    np.array([10]),
                ),
            ),
            (
                {"b": [12, 13], "a": (11,)},
                TrackedDict({"x": 14, "y": TrackedList([15, 16])}),
                TrackedSet([17]),
                Point(y=19, x=18),
                np.array([20]),
            ),
        )
        self.assertEqualStrict(
            v.visited(),
            [
                1,
                (11,),
                2,
                3,
                [12, 13],
                {"b": [12, 13], "a": (11,)},
                4,
                5,
                6,
                TrackedList([15, 16]),
                TrackedDict({"x": 14, "y": TrackedList([15, 16])}),
                7,
                TrackedSet([17]),
                8,
                9,
                Point(x=18, y=19),
                np.array([10]),
                (
                    {"b": [12, 13], "a": (11,)},
                    TrackedDict({"x": 14, "y": TrackedList([15, 16])}),
                    TrackedSet([17]),
                    Point(y=19, x=18),
                    np.array([20]),
                ),
            ],
        )

        # Error cases.
        with self.assertRaisesRegex(TypeError, "callable"):
            traverse_u("bad", [1, 2])

        # Children are not explored if structure is replaced with a leaf.
        v = Visitor(lambda x: "X" if isinstance(x, tuple) else None)
        self.assertEqualStrict(
            traverse_u(v, [(1, [2]), [3, (4, 5, 6)]]),
            ["X", [3, "X"]],
        )
        self.assertEqualStrict(
            v.visited(),
            [
                1,
                2,
                [2],
                (1, [2]),
                3,
                4,
                5,
                6,
                (4, 5, 6),
                [3, "X"],
                ["X", [3, "X"]],
            ],
        )

        # Children are not explored if structure is replaced with structure.
        v = Visitor(lambda x: ("a", "b") if isinstance(x, tuple) else None)
        self.assertEqualStrict(
            traverse_u(v, [(1, [2]), [3, (4, 5, 6)]]),
            [("a", "b"), [3, ("a", "b")]],
        )
        self.assertEqualStrict(
            v.visited(),
            [
                1,
                2,
                [2],
                (1, [2]),
                3,
                4,
                5,
                6,
                (4, 5, 6),
                [3, ("a", "b")],
                [("a", "b"), [3, ("a", "b")]],
            ],
        )

        # MAP_TO_NONE.
        v = Visitor(lambda x: MAP_TO_NONE if isinstance(x, tuple) else None)
        self.assertEqualStrict(
            traverse_u(v, [(1, [2]), [3, (4, 5, 6)]]),
            [None, [3, None]],
        )
        self.assertEqualStrict(
            v.visited(),
            [
                1,
                2,
                [2],
                (1, [2]),
                3,
                4,
                5,
                6,
                (4, 5, 6),
                [3, None],
                [None, [3, None]],
            ],
        )

    def test_lists_to_tuples(self, t):
        self.assertEqualStrict(
            t.lists_to_tuples([1, 2, 3]),
            (1, 2, 3),
        )
        self.assertEqualStrict(
            t.lists_to_tuples([[1], [2, 3]]),
            ((1,), (2, 3)),
        )

        # Deeper nested structures.
        self.assertEqualStrict(
            t.lists_to_tuples(
                (
                    {"b": [2, 3], "a": (1,)},
                    TrackedDict({"x": 4, "y": TrackedList([5, 6])}),
                    TrackedSet([(7, 8, 9)]),
                ),
            ),
            (
                {"b": (2, 3), "a": (1,)},
                TrackedDict({"x": 4, "y": (5, 6)}),
                TrackedSet([(7, 8, 9)]),
            ),
        )



# Modified from: keras/src/testing/test_case.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import shutil
import tempfile
import unittest

from absl.testing import parameterized
from dotenv import load_dotenv

from synalinks.src.backend.common.global_state import clear_session
from synalinks.src.backend.config import disable_telemetry


class TestCase(
    unittest.IsolatedAsyncioTestCase, parameterized.TestCase, unittest.TestCase
):
    maxDiff = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        # Load environment variables from .env file
        load_dotenv()
        # clear global state so that test cases are independent
        clear_session(free_memory=False)
        disable_telemetry()

    def get_temp_dir(self):
        temp_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(temp_dir))
        return temp_dir



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

"""
We provide different backend-dependent `DataModel`s to use.

These data models provide I/O for chatbots, agents, rags etc.

The user can build new data models by inheriting from these base models.

The check functions works for every type of data models (by checking the schema)
e.g. `SymbolicDataModel`, `JsonDataModel`, `DataModel` or `Variable`.
"""

from datetime import datetime
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from pydantic import Field
from pydantic import field_validator

from synalinks.src.api_export import synalinks_export
from synalinks.src.backend.common.json_schema_utils import contains_schema
from synalinks.src.backend.pydantic.core import DataModel


@synalinks_export(
    [
        "synalinks.backend.Score",
        "synalinks.Score",
    ]
)
class Score(float, Enum):
    VERY_BAD = 0.0
    POOR = 0.1
    BELOW_AVERAGE = 0.2
    LOW_AVERAGE = 0.3
    MEDIUM_LOW = 0.4
    MEDIUM = 0.5
    MEDIUM_HIGH = 0.6
    ABOVE_AVERAGE = 0.7
    HIGH_AVERAGE = 0.8
    GOOD = 0.9
    VERY_GOOD = 1.0


@synalinks_export(
    [
        "synalinks.backend.GenericOutputs",
        "synalinks.GenericOutputs",
    ]
)
class GenericOutputs(DataModel):
    """A generic outputs"""

    outputs: Dict[str, Any] = Field(
        description="The outputs",
    )


@synalinks_export(
    [
        "synalinks.backend.GenericInputs",
        "synalinks.GenericInputs",
    ]
)
class GenericInputs(DataModel):
    """A generic inputs"""

    inputs: Dict[str, Any] = Field(
        description="The inputs",
    )


@synalinks_export(
    [
        "synalinks.backend.Stamp",
        "synalinks.Stamp",
    ]
)
class Stamp(DataModel):
    created_at: Optional[datetime] = Field(
        description="The creation time",
        default=None,
    )


@synalinks_export(
    [
        "synalinks.backend.is_stamped",
        "synalinks.is_stamped",
    ]
)
def is_stamped(x):
    """Checks if the given data model is stamped

    Args:
        x (DataModel | JsonDataModel | SymbolicDataModel | Variable):
            The data model to check.

    Returns:
        (bool): True if the condition is met
    """
    schema = x.get_schema()
    properties = schema.get("properties", None)
    if properties:
        if properties.get("created_at", None):
            return True
    return False


@synalinks_export(
    [
        "synalinks.backend.GenericIO",
        "synalinks.GenericIO",
    ]
)
class GenericIO(DataModel):
    """A pair of generic inputs/outputs"""

    inputs: Dict[str, Any] = Field(
        description="The inputs",
    )
    outputs: Dict[str, Any] = Field(
        description="The outputs",
    )


@synalinks_export(
    [
        "synalinks.backend.GenericResult",
        "synalinks.GenericResult",
    ]
)
class GenericResult(DataModel):
    """A generic result"""

    result: List[Any] = Field(
        description="The result",
    )


@synalinks_export(
    [
        "synalinks.backend.ChatRole",
        "synalinks.ChatRole",
    ]
)
class ChatRole(str, Enum):
    """The chat message roles"""

    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"
    TOOL = "tool"


@synalinks_export(
    [
        "synalinks.backend.ToolCalling",
        "synalinks.ToolCalling",
        "synalinks.ToolCall",
        "synalinks.backend.ToollCall",
    ]
)
class ToolCall(DataModel):
    id: str = Field(
        description="The id of the tool call",
    )
    name: str = Field(
        description="The name of the function called",
    )
    arguments: Dict[str, Any] = Field(
        description="The arguments of the tool call",
    )


@synalinks_export(
    [
        "synalinks.backend.ChatMessage",
        "synalinks.ChatMessage",
    ]
)
class ChatMessage(DataModel):
    """A chat message"""

    role: ChatRole = Field(
        description="The chat message role",
    )
    content: Union[str, Dict[str, Any]] = Field(
        description="The content of the message",
        default="",
    )
    tool_call_id: Optional[str] = Field(
        description="The id of the tool call if role is `tool`",
        default=None,
    )
    tool_calls: List[ToolCall] = Field(
        description="The tool calls of the agent",
        default=[],
    )
    created_at: Optional[datetime] = Field(
        description="The creation time",
        default=None,
    )


@synalinks_export(
    [
        "synalinks.backend.is_chat_message",
        "synalinks.is_chat_message",
    ]
)
def is_chat_message(x):
    """Checks if the given data model is a chat message

    Args:
        x (DataModel | JsonDataModel | SymbolicDataModel | Variable):
            The data model to check.

    Returns:
        (bool): True if the condition is met
    """
    if contains_schema(x.get_schema(), ChatMessage.get_schema()):
        return True
    return False


@synalinks_export(
    [
        "synalinks.backend.ChatMessages",
        "synalinks.ChatMessages",
    ]
)
class ChatMessages(DataModel):
    """A list of chat messages"""

    messages: List[ChatMessage] = Field(
        description="The list of chat messages",
        default=[],
    )

    @field_validator("messages", mode="before")
    @classmethod
    def convert_dicts_to_chat_messages(cls, v):
        """Convert dict messages to ChatMessage objects."""
        if isinstance(v, list):
            return [ChatMessage(**msg) if isinstance(msg, dict) else msg for msg in v]
        return v


@synalinks_export(
    [
        "synalinks.backend.is_chat_messages",
        "synalinks.is_chat_messages",
    ]
)
def is_chat_messages(x):
    """Checks if the given data model are chat messages

    Args:
        x (DataModel | JsonDataModel | SymbolicDataModel | Variable):
            The data model to check.

    Returns:
        (bool): True if the condition is met
    """
    if contains_schema(x.get_schema(), ChatMessages.get_schema()):
        return True
    return False


@synalinks_export(
    [
        "synalinks.backend.is_tool_call",
        "synalinks.is_tool_call",
    ]
)
def is_tool_call(x):
    """Checks if the given data model is a tool call

    Args:
        x (DataModel | JsonDataModel | SymbolicDataModel | Variable):
            The data model to check.

    Returns:
        (bool): True if the condition is met
    """
    if contains_schema(x.get_schema(), ToolCall.get_schema()):
        return True
    return False


@synalinks_export(
    [
        "synalinks.backend.Embedding",
    ]
)
class Embedding(DataModel):
    """An embedding vector"""

    embedding: List[float] = Field(
        description="The embedding vector",
        default=[],
    )


@synalinks_export(
    [
        "synalinks.backend.is_embedding",
        "synalinks.is_embedding",
    ]
)
def is_embedding(x):
    """Checks if the given data model is an embedding

    Args:
        x (DataModel | JsonDataModel | SymbolicDataModel | Variable):
            The data model to check.

    Returns:
        (bool): True if the condition is met
    """
    if contains_schema(x.get_schema(), Embedding.get_schema()):
        return True
    return False


@synalinks_export(
    [
        "synalinks.backend.Embeddings",
        "synalinks.Embeddings",
    ]
)
class Embeddings(DataModel):
    """A list of embeddings"""

    embeddings: List[List[float]] = Field(
        description="The list of embedding vectors",
        default=[],
    )


@synalinks_export(
    [
        "synalinks.backend.is_embeddings",
        "synalinks.is_embeddings",
    ]
)
def is_embeddings(x):
    """Checks if the given data model are embeddings

    Args:
        x (DataModel | JsonDataModel | SymbolicDataModel | Variable):
            The data model to check.

    Returns:
        (bool): True if the condition is met
    """
    if contains_schema(x.get_schema(), Embeddings.get_schema()):
        return True
    return False


@synalinks_export(
    [
        "synalinks.backend.is_embedded",
        "synalinks.is_embedded",
    ]
)
def is_embedded(x):
    """Checks if the given data model is an embedded entity

    Args:
        x (DataModel | JsonDataModel | SymbolicDataModel | Variable):
            The data model to check.

    Returns:
        (bool): True if the condition is met
    """
    schema = x.get_schema()
    properties = schema.get("properties", None)
    if properties:
        if properties.get("embedding", None):
            return True
    return False


@synalinks_export(
    [
        "synalinks.backend.Prediction",
        "synalinks.Prediction",
    ]
)
class Prediction(GenericIO):
    reward: Optional[float] = Field(
        description="The prediction's reward",
        default=None,
    )


@synalinks_export(
    [
        "synalinks.backend.is_prediction",
        "synalinks.is_prediction",
    ]
)
def is_prediction(x):
    """Checks if the given data model is a prediction

    Args:
        x (DataModel | JsonDataModel | SymbolicDataModel | Variable):
            The data model to check.

    Returns:
        (bool): True if the condition is met
    """
    if contains_schema(x.get_schema(), Prediction.get_schema()):
        return True
    return False


@synalinks_export(
    [
        "synalinks.backend.Trainable",
        "synalinks.Trainable",
    ]
)
class Trainable(DataModel):
    examples: List[Prediction] = Field(
        description="The examples for few-shot learning",
        default=[],
    )
    current_predictions: List[Prediction] = Field(
        description="The current predictions store",
        default=[],
    )
    predictions: List[Prediction] = Field(
        description="The predictions store",
        default=[],
    )
    seed_candidates: List[Any] = Field(
        description="The seed candidates",
        default=[],
    )
    candidates: List[Any] = Field(
        description="The candidates",
        default=[],
    )
    best_candidates: List[Any] = Field(
        description="The best candidates",
        default=[],
    )
    history: List[Any] = Field(
        description="The candidates history",
        default=[],
    )
    nb_visit: int = Field(
        description="The number of visits",
        default=0,
    )
    cumulative_reward: float = Field(
        description="The cumulative reward",
        default=0.0,
    )


@synalinks_export(
    [
        "synalinks.backend.is_trainable",
        "synalinks.is_trainable",
    ]
)
def is_trainable(x):
    """Checks if the given data model is Trainable

    Args:
        x (DataModel | JsonDataModel | SymbolicDataModel | Variable):
            The data model to check.

    Returns:
        (bool): True if the condition is met
    """
    if contains_schema(x.get_schema(), Trainable.get_schema()):
        return True
    return False


@synalinks_export(
    [
        "synalinks.backend.Instructions",
        "synalinks.Instructions",
    ]
)
class Instructions(Trainable):
    """The instructions for the language model"""

    instructions: Optional[str] = Field(
        description="The instructions for the language model",
    )


@synalinks_export(
    [
        "synalinks.backend.is_instructions",
        "synalinks.is_instructions",
    ]
)
def is_instructions(x):
    """Checks if the given data model is an instructions data model

    Args:
        x (DataModel | JsonDataModel | SymbolicDataModel | Variable):
            The data model to check.

    Returns:
        (bool): True if the condition is met
    """
    if contains_schema(x.get_schema(), Instructions.get_schema()):
        return True
    return False


@synalinks_export(
    [
        "synalinks.backend.SimilaritySearch",
        "synalinks.SimilaritySearch",
    ]
)
class SimilaritySearch(DataModel):
    entity_label: str = Field(
        description=("The label of the entity to look for(use `*` to match them all)"),
    )
    similarity_search: str = Field(
        description=("The natural language similarity query to match specific entities"),
    )


@synalinks_export(
    [
        "synalinks.backend.is_similarity_search",
        "synalinks.is_similarity_search",
    ]
)
def is_similarity_search(x):
    """Checks if is a similarity search data model

    Args:
        x (DataModel | JsonDataModel | SymbolicDataModel | Variable):
            The data model to check.

    Returns:
        (bool): True if the condition is met
    """
    schema = x.get_schema()
    properties = schema.get("properties", None)
    if properties:
        if properties.get("entity_label", None) and properties.get(
            "similarity_search", None
        ):
            return True
    return False


@synalinks_export(
    [
        "synalinks.backend.TripletSearch",
        "synalinks.TripletSearch",
    ]
)
class TripletSearch(DataModel):
    subject_label: str = Field(
        description=("The label of the subject entity (use `*` to match them all)"),
    )
    subject_similarity_search: str = Field(
        description=(
            "A short similarity query to match specific subjects"
            "(use `*` to match them all)",
        ),
    )
    relation_label: str = Field(
        description="The label of the relation to search for",
    )
    object_label: str = Field(
        description=("The label of the object entity (use `*` to match them all)"),
    )
    object_similarity_search: str = Field(
        description=(
            "A short similarity query to match specific objects"
            " (use `*` to match them all)"
        ),
    )


@synalinks_export(
    [
        "synalinks.backend.is_triplet_search",
        "synalinks.is_triplet_search",
    ]
)
def is_triplet_search(x):
    """Checks if is a triplet seach data model

    Args:
        x (DataModel | JsonDataModel | SymbolicDataModel | Variable):
            The data model to check.

    Returns:
        (bool): True if the condition is met
    """
    schema = x.get_schema()
    properties = schema.get("properties", None)
    if properties:
        if (
            properties.get("subject_label", None)
            and properties.get("subject_similarity_search", None)
            and properties.get("relation_label", None)
            and properties.get("object_label", None)
            and properties.get("object_similarity_search", None)
        ):
            return True
    return False



# Modified from: keras/src/trainers/data_adapters/generator_data_adapter.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import itertools

from synalinks.src.trainers.data_adapters import data_adapter_utils
from synalinks.src.trainers.data_adapters.data_adapter import DataAdapter


class GeneratorDataAdapter(DataAdapter):
    """Adapter for Python generators."""

    def __init__(self, generator):
        first_batches, generator = peek_and_restore(generator)
        self.generator = generator
        self._first_batches = first_batches
        self._output_signature = None
        if not isinstance(first_batches[0], tuple):
            raise ValueError(
                "When passing a Python generator to a Synalinks program, "
                "the generator must return a tuple, either "
                "(input,) or (inputs, targets). "
                f"Received: {first_batches[0]}"
            )

    def get_numpy_iterator(self):
        return data_adapter_utils.get_numpy_iterator(self.generator())

    @property
    def num_batches(self):
        return None

    @property
    def batch_size(self):
        return None


def peek_and_restore(generator):
    batches = list(itertools.islice(generator, data_adapter_utils.NUM_BATCHES_FOR_SPEC))
    return batches, lambda: itertools.chain(batches, generator)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import tree
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import JsonDataModel
from synalinks.src.backend import SymbolicDataModel
from synalinks.src.modules.module import Module


@synalinks_export(["synalinks.modules.Identity", "synalinks.Identity"])
class Identity(Module):
    """Identity module.

    This module should be used as a placeholder when no operation is to be
    performed. The module just returns its `inputs` argument as output.

    This module can be really useful during development process in order to
    implement the whole program architecture before the individual modules.

    It avoid any data models naming issue that you could have by just
    forwarding the inputs, that way you can implement the general
    program architecture, validate it and implement the individual
    modules later.

    Example:

    ```python
    import synalinks

    class MyAwesomeModule(synalinks.Program):

        def __init__(
            name=None,
            description=None,
            trainable=True,
        ):
            super().__init__(
                name=name,
                description=description,
                trainable=trainable,
            )

        async def build(self, inputs):
            outputs = await synalinks.Identity()(inputs)

            super().__init__(
                inputs=inputs,
                outputs=outputs,
                name=self.name,
                description=self.description,
                trainable=self.trainable,
            )
    ```

    Args:
        **kwargs (keyword arguments): The default module's arguments
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.built = True

    async def call(self, inputs):
        if isinstance(inputs, (JsonDataModel, SymbolicDataModel)):
            return inputs.clone()
        return tree.map_structure(
            lambda x: x.clone(),
            inputs,
        )



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.optimizers.evolutionary_optimizer import (
    EvolutionaryOptimizer as EvolutionaryOptimizer,
)
from synalinks.src.optimizers.greedy_optimizer import GreedyOptimizer as GreedyOptimizer
from synalinks.src.optimizers.omega import OMEGA as OMEGA
from synalinks.src.optimizers.random_few_shot import RandomFewShot as RandomFewShot



# Modified from: keras/src/backend/common/variables.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import backend
from synalinks.src.backend.common import global_state
from synalinks.src.backend.common.json_data_model import JsonDataModel
from synalinks.src.backend.common.json_schema_utils import standardize_schema
from synalinks.src.backend.common.name_scope import current_path
from synalinks.src.backend.common.stateless_scope import get_stateless_scope
from synalinks.src.backend.common.stateless_scope import in_stateless_scope
from synalinks.src.backend.common.symbolic_data_model import SymbolicDataModel
from synalinks.src.utils.naming import auto_name


class Variable:
    """A backend-agnostic variable in synalinks.

    A `Variable` acts as a container for state.
    It holds a JSON object value with the corresponding schema and can
    be updated by the optimizers.

    A Variable is different from a JsonDataModel as it can be modified by the optimizers.

    Note that the DataModel used for the variable declaration
    **must have a default value** for each of its field.

    Examples:

    **Initializing a `Variable` with a dict:**

    ```python
    from typing import List
    import synalinks

    class Instructions(synalinks.DataModel):
        instructions: List[str] = []

    initial_data = {
        "instructions": [
            "For any problem involving division, always round the quotient to "
            "the nearest even number, regardless of the remainder."
        ],
    }
    variable_from_dict = synalinks.Variable(
        initializer=initial_data,
        data_model=Instructions,
    )
    ```

    **Using a synalinks initializer to create a `Variable`:**

    ```python
    from typing import List
    import synalinks

    class Instructions(synalinks.DataModel):
        instructions: List[str] = []

    from synalinks.initializers import Empty

    variable_from_initializer = synalinks.Variable(
        initializer=Empty(data_model=Instructions)
    )
    ```

    **Updating the value of a `Variable`:**

    ```python
    new_json = {
        "instructions": [
            "When performing division, always check if the division results "
            "in a whole number. If not, express the result as a fraction or "
            "a decimal, depending on the context of the problem."
        ],
    }
    variable_from_dict.assign(new_json)
    ```

    **Marking a `Variable` as non-trainable:**

    ```python
    from typing import List
    import synalinks

    class Instructions(synalinks.DataModel):
        instructions: List[str] = []

    from synalinks.initializers import Empty

    non_trainable_variable = synalinks.Variable(
        initializer=Empty(data_model=Instructions), trainable=False
    )
    ```

    Args:
        initializer (str | dict | Initializer): Initial value (dict) or callable
            (Initializer) for initialization. If a callable is used, it should
            take the arguments `data_model`.
        data_model (DataModel): The backend-dependent data model used as spec.
        trainable (bool): Optional. Boolean indicating if variable is trainable.
            Defaults to `True`.
        name (str): Optional. A unique name for the variable. Automatically
            generated if not set.
        description (str): Optional. The description of the variable.
            Automatically generated if not set by fetching the data model
            description.
    """

    def __init__(
        self,
        initializer=None,
        data_model=None,
        trainable=True,
        name=None,
        description=None,
    ):
        name = name or auto_name(self.__class__.__name__)
        if not isinstance(name, str) or "/" in name:
            raise ValueError(
                "Argument `name` must be a string and "
                "cannot contain character `/`. "
                f"Received: name={name}"
            )
        self._name = name
        parent_path = current_path()
        if parent_path:
            self._path = current_path() + "/" + name
        else:
            self._path = name

        self._initializer = None
        self._data_model = data_model
        self._trainable = bool(trainable)
        self._json = None

        if in_stateless_scope():
            if callable(initializer):
                self._initializer = initializer
                self._schema = standardize_schema(data_model.get_schema())
                register_uninitialized_variable(self)
            else:
                raise ValueError(
                    "You are attempting to create a variable "
                    "while in a stateless scope. This is disallowed. "
                    "Make sure that all variables are created "
                    "before you start using your module/program objects.\n\n"
                    "In some cases, you might be seeing this error "
                    "because you need to "
                    "implement a `def build(self, input_schema)` method "
                    "on your module/program, which will "
                    "create its variables.\n\n"
                    "In some other cases, you might be seeing this error "
                    "because you are instantiating a `Variable` and "
                    "assigning it to a module without going through "
                    "self.add_variable(). Always prefer "
                    "using these methods "
                    "(with a `data_model` and `initializer` argument)."
                )
        else:
            if callable(initializer):
                self._initialize_with_initializer(initializer)
            else:
                if data_model is None:
                    raise ValueError(
                        "When creating a Variable from an a dict,"
                        "the `data_model` argument should be specified. "
                        f"Received: initializer={initializer} "
                        f"and data_model={data_model}"
                    )
                value = initializer
                self._initialize(value)
                self._schema = standardize_schema(data_model.get_schema())

        if not description:
            if "description" in self._schema:
                self._description = self._schema["description"]
            else:
                self._description = ""
        else:
            self._description = description

    def _deferred_initialize(self):
        """Deferred initialization of the variable.

        Raises:
            ValueError: If the variable is already initialized or
                if attempting to initialize while in a stateless scope.
        """
        if self._json is not None:
            raise ValueError(f"Variable '{self._path}' is already initialized.")

        if in_stateless_scope():
            raise ValueError(
                "You are attempting to initialize a variable "
                "while in a stateless scope. This is disallowed. "
                "Make sure that all variables are initialized "
                "before you start using your module/program objects."
            )
        self._initialize_with_initializer(self._initializer)
        self._initializer = None

    def get_json(self):
        """The current value of the variable.

        Returns:
            (dict): The current value of the variable.
        """
        if in_stateless_scope():
            scope = get_stateless_scope()
            value = scope.get_current_value(self)
            if value is not None:
                return value
        if self._json is None:
            # Uninitialized variable. Return a placeholder.
            # This is fine because it's only ever used
            # in during schema inference / graph tracing
            # (anything else would be a bug, to be fixed.)
            return self._initializer(data_model=self._data_model)
        return self._json

    def prettify_schema(self):
        """Get a pretty version of the JSON schema for display.

        Returns:
            (dict): The indented JSON schema.
        """
        import orjson

        return orjson.dumps(self.get_schema(), option=orjson.OPT_INDENT_2).decode()

    def prettify_json(self):
        """Get a pretty version of the JSON object for display.

        Returns:
            (dict): The indented JSON object.
        """
        import orjson

        return orjson.dumps(self.get_json(), option=orjson.OPT_INDENT_2).decode()

    def assign(self, value):
        """Assigns a new value to the variable.

        Args:
            value (dict | DataModel | JsonDataModel): The new value to be assigned.
                The value can be an instanciated data model or JSON dict.

        Returns:
            (dict): The assigned value.

        Raises:
            ValueError: If the schema of the target variable and
                the value are incompatible.
        """
        if backend.is_data_model(value):
            value = value.get_json()
        if in_stateless_scope():
            scope = get_stateless_scope()
            scope.add_update((self, value))
        else:
            self._direct_assign(value)
        return value

    def _direct_assign(self, json):
        """Directly assigns a new value to the variable.

        Args:
            json (dict): The new json value to be assigned.
        """
        self._json = json

    def get_schema(self):
        """The schema of the variable.

        Returns:
            (dict): The JSON schema of the variable.
        """
        return self._schema

    def __contains__(self, other):
        """Check if the schema of `other` is contained within the schema of this variable,
        or if a string key exists.

        Args:
            other (SymbolicDataModel | DataModel | str): The data model to compare
                against this variable's schema, or a string key to check for in the
                schema properties.

        Returns:
            (bool): True if `other`'s schema is a subset of this variable's schema,
                or if the string key exists in the schema properties.
        """
        if isinstance(other, str):
            schema = self.get_schema()
            return other in schema.get("properties", {})
        from synalinks.src.backend.common.json_schema_utils import contains_schema

        return contains_schema(self.get_schema(), other.get_schema())

    @property
    def trainable(self):
        """Whether the variable is trainable."""
        return self._trainable

    @trainable.setter
    def trainable(self, value):
        self._trainable = bool(value)

    @property
    def name(self):
        """The name of the variable."""
        return self._name

    @property
    def description(self):
        """The description of the variable."""
        return self._description

    @property
    def path(self):
        """The path of the variable within the program or module."""
        return self._path

    def __repr__(self):
        json = None
        if self._json is not None:
            json = self._json
        json_str = f", json={json}" if json is not None else ""
        return f"<Variable path={self.path}, schema={self._schema}{json_str}>"

    def _initialize_with_initializer(self, initializer):
        """Initializes the variable using an initializer object.

        Args:
            initializer (Initializer): The initializer to be used.
        """
        value = initializer()
        self._schema = standardize_schema(initializer.get_schema())
        self._initialize(value)

    def _initialize(self, json):
        """Initializes the variable with a given json dict.

        Args:
            json (dict): The initial value (JSON object dict).
        """
        self._json = json

    def to_json_data_model(self, name=None):
        """Convert the variable into a `JsonDataModel`.

        Returns:
            (JsonDataModel): The equivalent backend-independent data model
        """
        if not name:
            name = self.name
        return JsonDataModel(
            schema=self.get_schema(),
            json=self.get_json(),
            name=name,
        )

    def to_symbolic_data_model(self, name=None):
        """Convert the variable into a `SymbolicDataModel`.

        Returns:
            (SymbolicDataModel): The equivalent symbolic data model
        """
        if not name:
            name = self.name
        return SymbolicDataModel(
            schema=self.get_schema(),
            name=name,
        )

    def get(self, key, default=None):
        """Get wrapper to make it easier to access JSON fields.

        Args:
            key (str): The key to access.
            default (any): The default value if key not found.
        """
        return self.get_json().get(key, default)

    def __getitem__(self, key):
        """Get item wrapper to make it easier to access JSON fields.

        Args:
            key (str): The key to access.
        """
        return self.get_json()[key]

    def keys(self):
        """Keys wrapper to make it easier to access JSON fields."""
        return self.get_json().keys()

    def values(self):
        """Values wrapper to make it easier to access JSON fields."""
        return self.get_json().values()

    def items(self):
        """Items wrapper to make it easier to access JSON fields."""
        return self.get_json().items()

    def update(self, kv_dict):
        """Update wrapper to make easier to modify fields.

        Args:
            kv_dict (dict): The key/value dict to update.
        """
        self._json.update(kv_dict)


def register_uninitialized_variable(variable):
    uninitialized_variables = global_state.get_global_attribute(
        "uninitialized_variables", [], set_to_default=True
    )
    uninitialized_variables.append(variable)


def initialize_all_variables():
    collection = global_state.get_global_attribute("uninitialized_variables")
    if collection:
        for v in collection:
            v._deferred_initialize()
    global_state.set_global_attribute("uninitialized_variables", [])



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from unittest.mock import AsyncMock
from unittest.mock import MagicMock

from synalinks.src import testing
from synalinks.src.backend import JsonDataModel
from synalinks.src.optimizers.evolutionary_optimizer import EvolutionaryOptimizer
from synalinks.src.optimizers.omega import OMEGA
from synalinks.src.optimizers.omega import base_instructions
from synalinks.src.optimizers.omega import crossover_instructions
from synalinks.src.optimizers.omega import mutation_instructions
from synalinks.src.optimizers.omega import similarity_distance


class OMEGATest(testing.TestCase):
    def test_inheritance(self):
        """Test that OMEGA inherits from EvolutionaryOptimizer."""
        self.assertTrue(issubclass(OMEGA, EvolutionaryOptimizer))

    def test_init_default_parameters(self):
        """Test initialization with default parameters."""
        optimizer = OMEGA()

        self.assertIsNone(optimizer.language_model)
        self.assertIsNone(optimizer.embedding_model)
        self.assertEqual(optimizer.mutation_temperature, 0.3)
        self.assertEqual(optimizer.crossover_temperature, 0.3)
        self.assertEqual(optimizer.k_nearest_fitter, 5)
        self.assertEqual(optimizer.algorithm, "dns")
        self.assertEqual(optimizer.selection, "softmax")
        self.assertEqual(optimizer.selection_temperature, 0.3)
        self.assertEqual(optimizer.merging_rate, 0.02)
        self.assertEqual(optimizer.population_size, 10)
        self.assertEqual(optimizer.instructions, "")

    def test_init_custom_parameters(self):
        """Test initialization with custom parameters."""
        mock_lm = MagicMock()
        mock_em = MagicMock()

        optimizer = OMEGA(
            instructions="Test instructions",
            language_model=mock_lm,
            embedding_model=mock_em,
            mutation_temperature=0.5,
            crossover_temperature=0.7,
            k_nearest_fitter=10,
            algorithm="ga",
            selection="best",
            selection_temperature=0.2,
            merging_rate=0.1,
            population_size=20,
            name="test_omega",
            description="Test OMEGA optimizer",
        )

        self.assertEqual(optimizer.language_model, mock_lm)
        self.assertEqual(optimizer.embedding_model, mock_em)
        self.assertEqual(optimizer.mutation_temperature, 0.5)
        self.assertEqual(optimizer.crossover_temperature, 0.7)
        self.assertEqual(optimizer.k_nearest_fitter, 10)
        self.assertEqual(optimizer.algorithm, "ga")
        self.assertEqual(optimizer.selection, "best")
        self.assertEqual(optimizer.selection_temperature, 0.2)
        self.assertEqual(optimizer.merging_rate, 0.1)
        self.assertEqual(optimizer.population_size, 20)
        self.assertEqual(optimizer.instructions, "Test instructions")

    def test_invalid_algorithm(self):
        """Test that invalid algorithm raises ValueError."""
        with self.assertRaises(ValueError) as context:
            OMEGA(algorithm="invalid")

        self.assertIn("algorithm", str(context.exception))

    def test_valid_algorithms(self):
        """Test that valid algorithms are accepted."""
        for algorithm in ["ga", "dns"]:
            optimizer = OMEGA(algorithm=algorithm)
            self.assertEqual(optimizer.algorithm, algorithm)

    def test_get_config(self):
        """Test that get_config returns all parameters."""
        optimizer = OMEGA(
            instructions="Test",
            mutation_temperature=0.4,
            crossover_temperature=0.6,
            k_nearest_fitter=8,
            algorithm="ga",
            selection="random",
            selection_temperature=0.25,
            merging_rate=0.05,
            population_size=15,
            name="config_test",
            description="Config test optimizer",
        )

        config = optimizer.get_config()

        self.assertEqual(config["instructions"], "Test")
        self.assertEqual(config["mutation_temperature"], 0.4)
        self.assertEqual(config["crossover_temperature"], 0.6)
        self.assertEqual(config["k_nearest_fitter"], 8)
        self.assertEqual(config["algorithm"], "ga")
        self.assertEqual(config["selection"], "random")
        self.assertEqual(config["selection_temperature"], 0.25)
        self.assertEqual(config["merging_rate"], 0.05)
        self.assertEqual(config["population_size"], 15)

    async def test_competition_single_candidate(self):
        """Test competition with single candidate returns it unchanged."""
        optimizer = OMEGA()

        candidates = [{"prompt": "test", "reward": 0.8}]
        result = await optimizer.competition(candidates)

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], candidates[0])

    async def test_competition_empty_candidates(self):
        """Test competition with empty candidates returns empty list."""
        optimizer = OMEGA()

        result = await optimizer.competition([])

        self.assertEqual(result, [])

    async def test_competition_filters_candidates(self):
        """Test that competition filters candidates based on DNS."""
        # Create mock embedding model
        mock_embedding_model = AsyncMock()
        mock_embedding_model.return_value = {
            "embeddings": [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]]
        }

        optimizer = OMEGA(
            embedding_model=mock_embedding_model,
            k_nearest_fitter=2,
        )

        # Create candidates with different rewards
        candidates = [
            {"prompt": "test1", "reward": 0.9},
            {"prompt": "test2", "reward": 0.5},
            {"prompt": "test3", "reward": 0.7},
            {"prompt": "test4", "reward": 0.3},
        ]

        result = await optimizer.competition(candidates)

        # Should filter out some candidates based on DNS
        self.assertGreater(len(result), 0)
        self.assertLessEqual(len(result), len(candidates))

    async def test_on_epoch_end_sorts_and_selects_candidates_ga(self):
        """Test on_epoch_end sorts candidates and selects top ones (GA mode)."""
        optimizer = OMEGA(
            algorithm="ga",  # Skip DNS competition
            population_size=2,
        )

        candidates = [
            {"prompt": "c1", "reward": 0.3},
            {"prompt": "c2", "reward": 0.9},
        ]
        best_candidates = [
            {"prompt": "b1", "reward": 0.5},
        ]

        trainable_variable = JsonDataModel(
            json={
                "candidates": candidates,
                "best_candidates": best_candidates,
            },
            schema={
                "type": "object",
                "properties": {
                    "candidates": {"type": "array"},
                    "best_candidates": {"type": "array"},
                },
            },
        )

        await optimizer.on_epoch_end(0, [trainable_variable])

        # Should keep top 2 candidates by reward
        result = trainable_variable.get("best_candidates")
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0]["reward"], 0.9)
        self.assertEqual(result[1]["reward"], 0.5)

    async def test_on_epoch_end_with_dns(self):
        """Test on_epoch_end applies DNS competition when algorithm='dns'."""
        # Create mock embedding model
        mock_embedding_model = AsyncMock()
        mock_embedding_model.return_value = {"embeddings": [[0.1, 0.2, 0.3]]}

        optimizer = OMEGA(
            algorithm="dns",
            embedding_model=mock_embedding_model,
            population_size=3,
        )

        candidates = [
            {"prompt": "c1", "reward": 0.9},
            {"prompt": "c2", "reward": 0.5},
        ]
        best_candidates = [
            {"prompt": "b1", "reward": 0.7},
        ]

        trainable_variable = JsonDataModel(
            json={
                "candidates": candidates,
                "best_candidates": best_candidates,
            },
            schema={
                "type": "object",
                "properties": {
                    "candidates": {"type": "array"},
                    "best_candidates": {"type": "array"},
                },
            },
        )

        await optimizer.on_epoch_end(0, [trainable_variable])

        # Should have applied DNS and sorted
        result = trainable_variable.get("best_candidates")
        self.assertGreater(len(result), 0)
        self.assertLessEqual(len(result), 3)


class SimilarityDistanceTest(testing.TestCase):
    async def test_similarity_distance_identical_candidates(self):
        """Test similarity_distance returns 0 for identical candidates."""
        mock_embedding_model = AsyncMock()
        mock_embedding_model.return_value = {"embeddings": [[1.0, 0.0, 0.0]]}

        candidate = {"prompt": "test"}

        distance = await similarity_distance(
            candidate, candidate, embedding_model=mock_embedding_model
        )

        # Identical embeddings should have distance close to 0
        self.assertLessEqual(distance, 0.1)

    async def test_similarity_distance_different_candidates(self):
        """Test similarity_distance for different candidates."""
        call_count = [0]

        async def mock_embed(texts):
            call_count[0] += 1
            if call_count[0] == 1:
                return {"embeddings": [[1.0, 0.0, 0.0]]}
            else:
                return {"embeddings": [[0.0, 1.0, 0.0]]}

        mock_embedding_model = AsyncMock(side_effect=mock_embed)

        candidate1 = {"prompt": "test1"}
        candidate2 = {"prompt": "test2"}

        distance = await similarity_distance(
            candidate1, candidate2, embedding_model=mock_embedding_model
        )

        # Orthogonal embeddings should have distance around 0.5
        self.assertGreater(distance, 0.3)
        self.assertLess(distance, 0.7)


class InstructionFunctionsTest(testing.TestCase):
    def test_base_instructions(self):
        """Test base_instructions returns non-empty string."""
        result = base_instructions()
        self.assertIsInstance(result, str)
        self.assertGreater(len(result), 0)
        self.assertIn("optimization", result.lower())

    def test_mutation_instructions(self):
        """Test mutation_instructions includes variable keys."""
        keys = ["prompt", "rules", "examples"]
        result = mutation_instructions(keys)

        self.assertIsInstance(result, str)
        self.assertIn(str(keys), result)
        self.assertIn("enhance", result.lower())

    def test_crossover_instructions(self):
        """Test crossover_instructions includes variable keys."""
        keys = ["prompt", "rules"]
        result = crossover_instructions(keys)

        self.assertIsInstance(result, str)
        self.assertIn(str(keys), result)
        self.assertIn("combining", result.lower())



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import tree
from synalinks.src.api_export import synalinks_export
from synalinks.src.modules.module import Module


@synalinks_export(
    [
        "synalinks.InMask",
        "synalinks.modules.InMask",
    ]
)
class InMask(Module):
    """A module to keep specific fields of the given data models

    Example:

    ```python
    import synalinks
    import asyncio

    language_model = synalinks.LanguageModel(
        model="ollama/mistral",
    )

    class Document(synalinks.DataModel):
        title: str = synalinks.Field(
            description="The title of the document",
        )
        text: str = synalinks.Field(
            description="The content of the document",
        )

    class Summary(synalinks.DataModel):
        summary: str = synalinks.Field(
            description="the concise summary of the document",
        )

    async def main():
        inputs = Input(data_model=Document)
        summary = synalinks.ChainOfThought(
            data_model=Summary,
            language_model=language_model,
        )(inputs)
        masked_summary = synalinks.InMask(
            # remove the thinking field from the chain of thought
            # by keeping only the summary
            mask=["summary"],
        )(summary)

        program = Program(
            inputs=inputs,
            outputs=masked_summary,
            name="summary_generator",
            description="Generate a summary from a document",
        )

    ```

    Args:
        mask (list): The list of keys to keep.
        name (str): Optional. The name of the module.
        description (str): Optional. The description of the module.
        trainable (bool): Whether the module's variables should be trainable.
    """

    def __init__(
        self,
        mask=None,
        name=None,
        description=None,
        trainable=False,
    ):
        if not mask or not isinstance(mask, list):
            raise ValueError("`mask` parameter should be a list of fields to keep")
        super().__init__(
            name=name,
            description=description,
        )
        self.mask = mask

    async def call(self, inputs):
        outputs = tree.map_structure(
            lambda x: x.in_mask(mask=self.mask),
            inputs,
        )
        return outputs
